"""DearPyGui item error utilities."""
import inspect
import functools
from dearpygui import _dearpygui
from ._typing import Item, ItemCommand, Protocol, ParamSpec
from . import _parsing


_P = ParamSpec("_P")


class ItemError(Protocol[_P]):
    def __call__(self, item: Item, command: ItemCommand[_P, Item]) -> Exception | None: ...


# [ HELPER FUNCTIONS ]

def is_tag_valid(tag: Item) -> bool:
    return isinstance(tag, (int, str))


def is_item_interface(item: Item) -> bool:
    if type(item) in (str, int):
        return False
    elif is_tag_valid(item):
        return True
    return False


def does_item_exist(tag: Item) -> bool:
    try:
        if isinstance(tag, int):
            return _dearpygui.does_item_exist(tag)
        elif isinstance(tag, str):
            return _dearpygui.does_alias_exist(tag)
    except SystemError:
        return False
    return False


def command_signature(command: ItemCommand):
    variadics = (inspect.Parameter.VAR_KEYWORD, inspect.Parameter.VAR_POSITIONAL)
    signature = inspect.signature(command)
    return signature.replace(
        parameters=tuple(
            p for p in signature.parameters.values()
            if p.kind not in variadics
        )
    )


def command_itp_name(command: ItemCommand) -> str:
    from . import api

    for tp_def in _parsing.item_definitions().values():
        try:
            if command == tp_def.command1 or command == tp_def.command2:  # type: ignore
                return tp_def.name  # type: ignore
        except AttributeError:
            pass
    return '<unknown type>'




# [ ERROR FUNCTIONS ]

def err_tag_invalid(item: Item, command: ItemCommand, *args, **kwargs):
    if not is_tag_valid(item):
        return TypeError(f'`tag` expected as int or str, got {type(item)!r}.')


def err_item_nonexistant(item: Item, command: ItemCommand, *args, **kwargs):
    if isinstance(item, int):
        item_exists = _dearpygui.does_item_exist(item)
        id_type     = 'uuid'
    elif isinstance(item, str):
        item_exists = _dearpygui.does_alias_exist(item)
        id_type     = 'alias'
    else:
        return err_tag_invalid(item, command, *args, **kwargs)
    if item and item_exists:
        return ValueError(f"item {id_type} already exists and/or is in use.")


def err_arg_unexpected(item: Item, command: ItemCommand, *args, **kwargs):
    signature = command_signature(command)
    # Python will automatically throw a TypeError for unexpected
    # positional args since `dearpygui.dearpygui` is valid code.
    # All item commands accept variadic keywords though, and
    # `mvPythonParser` will throw vague errors for them.
    bad_kwds = ', '.join(repr(k) for k in kwargs if k not in signature.parameters)
    if not bad_kwds:
        return
    if not is_tag_valid(item):
        return err_tag_invalid(item, command, *args, **kwargs)
    elif is_item_interface(item):
        callee = f'{type(item).__qualname__}()'
    else:
        callee = f'{command.__qualname__}()'  # type: ignore

    return TypeError(f'{callee!r} received unexpected keyword argument(s) {bad_kwds}.')


def err_arg_invalid(item: Item, command: ItemCommand, *args, **kwargs):
    # TODO
    ...





def _invld_p_err(parent_tp: str, child_tp: str, *tp_list: str) -> ValueError:
    msg = f"incompatible parent {parent_tp!r} for {child_tp!r} item{{}}."
    if tp_list:
        if isinstance(tp_list, str):
            tp_list = ''.join(tp_list),
        msg = msg.format(f", try {', '.join(repr(s) for s in tp_list)}.")
    else:
        msg = msg.format(".")
    return ValueError(msg)

def err_parent_invalid(item: Item, command: ItemCommand, *args, **kwargs):
    if "parent" not in command_signature(command).parameters:
        return

    cstack = _dearpygui.top_container_stack()
    parent = kwargs.get("parent", 0)

    # CASE 1: `parent` is not `int` nor `str` (also prevents error in CASE 2)
    if not is_tag_valid(parent):
        return ValueError(f"`parent` expected as int or str, got {type(parent)!r}.")

    if parent:
        # CASE 2: the included parent does not exist
        if not does_item_exist(parent):
            return ValueError(f"`parent` item {parent!r} does not exist.")
        # CASE 3: is it even a container?
        if _dearpygui.get_item_info(parent)['container'] is False:
            raise ValueError(f"`parent` item {parent!r} is not a container and cannot parent items.")

    # CASE 4: did not include parent while the container stack is empty
    elif not parent and not cstack:
        return TypeError("no parent item available (container stack is empty).")


    parent = parent or cstack

    # CASE 5: a parent is available but maybe incompatible
    p_tp_name = _dearpygui.get_item_info(
        parent
    )["type"].removeprefix("mvAppItemType::")
    c_tp_name = command_itp_name(command).removeprefix('mvAppItemType::')

    # these are supposed to be "universal" parents, but there are some exceptions
    if p_tp_name in ("mvStage", "mvTemplateRegistry"):
        return ValueError(f"incompatible parent {p_tp_name!r} for {c_tp_name!r} item.")

    err = functools.partial(_invld_p_err, p_tp_name, c_tp_name)

    if c_tp_name.startswith("mvDraw") and p_tp_name not in ("mvDraw", "mvViewportDraw"):
        return err('mvDrawlist', 'mvViewportDrawlist', 'mvDrawLayer')
    if c_tp_name.startswith("mvThemeComponent") and p_tp_name not in ("mvTheme",):
        return err('mvTheme',)
    if c_tp_name in ("mvThemeColor", 'mvThemeStyle') and p_tp_name not in ("mvThemeComponent",):
        return err("mvThemeComponent",)
    if "Handler" in c_tp_name:
        if p_tp_name == "mvHandlerRegistry":
            return err("mvItemHandlerRegistry",)
        if p_tp_name == "mvItemHandlerRegistry":
            return err("mvItemHandlerRegistry",)
    if c_tp_name.endswith("Series") and p_tp_name != 'mvPlotAxis':
        return err('mvPlotAxis',)
    if c_tp_name in ('mvAnnotation', 'mvPlotLegend', 'mvPlotAxis') and p_tp_name != "mvPlot":
        return err("mvPlot",)
    if c_tp_name.startswith('mvNode'):
        if c_tp_name == 'mvNodeAttribute' and p_tp_name != 'mvNode':
            return err('mvNode',)
        if c_tp_name == 'mvNode' and p_tp_name != 'mvNodeEditor':
            return err('mvNodeEditor',)
    if c_tp_name in ('mvTableRow', 'mvTableCell') and p_tp_name != 'mvTable':
        return err('mvTable',)
    if c_tp_name == 'mvTableCell' and p_tp_name != 'mvTableRow':
        return err('mvTableRow',)





from dearpygui import dearpygui
from ._typing import (
    Item,
    ItemUUID,
    ItemAlias,
    Interface,
    ItemCallback,
    ItemCommand,
    ItemInfoDict,
    ItemStateDict,

    mvBuffer,
    mvMat4,
    mvVec4,

    cast,
)
from ._interface import (
    AppItemMeta,
    AppItemType,

    ABCAppItemMeta,
    ABCAppItemType,

    BasicType,
    ContainerType,
    RootType,
    RegistryType,

    HandlerType,
    NodeType,
    NodeEditorType,
    ThemeType,
    ThemeElementType,
    FontType,
    DrawingType,
    DrawNodeType,
    PlottingType,
    PlotType,
    PlotAxisType,
    TableType,
    TableItemType,
    WindowType,

    SupportsCallback,
    SupportsSized,
    SupportsValueArray,
)
from .api import (
    _AppConfigDict as AppConfigDict,
    _ViewportConfigDict as ViewportConfigDict,
    _RuntimeConfigDict as RuntimeConfigDict,
)


mvBuffer = cast(type[mvBuffer], dearpygui.mvBuffer)
mvMat4   = cast(type[mvMat4], dearpygui.mvMat4)
mvVec4   = cast(type[mvVec4], dearpygui.mvVec4)



import sys
import numbers
import functools
import threading
import dataclasses
import math
from array import array
import dearpygui.dearpygui as dearpygui
from dearpygui._dearpygui import (
    get_item_configuration as _item_get_config,
    get_item_state as _item_get_state,
    configure_item as _item_set_config,
)
from typing import (
    Any,

    Self,
    Protocol,
    Callable,
    ClassVar,
    Literal,
    Sequence,
    Generic,
    Iterator,
    SupportsIndex,
    Sized,
    Iterable,
    Annotated,
    TypeVar,
    Union,

    overload,
    cast,
)




_T  = TypeVar('_T')
_N  = TypeVar('_N', bound=numbers.Real)

Item   = int | str
FloatV = float | None
Vec2   = Annotated[tuple[FloatV, FloatV]                 | Sequence[FloatV], Literal[2]]
Vec4   = Annotated[tuple[FloatV, FloatV, FloatV, FloatV] | Sequence[FloatV], Literal[4]]

NaN  = math.nan




# [ HELPER FUNCTIONS ]

def _create_context():
    # runs once, then no-op
    dearpygui.create_context()
    @functools.wraps(_create_context)
    def create_context(): ...
    setattr(sys.modules[__name__], _create_context.__name__, create_context)


def _is_nan(v: Any) -> bool:
    return v != v

def _is_nanlike(v: Any) -> bool:
    return v is None or v != v

def _to_value(value: Any, default: Any):
    if value is None or _is_nan(value):
        return default
    return value

def _to_float_arr(value: Sequence[float | None] | float | None, length: int, default: float = NaN) -> 'array[float]':
    """
    # Non-Sequence value (None | float('nan'))
    >>> arr = _to_float_array(None, 2)
    >>> len(arr) == 2 and all(_is_nan(v) for v in arr)
    True
    >>> arr = _to_float_array(float("nan"), 2)
    >>> len(arr) == 2 and all(_is_nan(v) for v in arr)
    True

    # Non-Sequence value (number)
    >>> arr = _to_float_array(20, 4)
    >>> len(arr) == 4 and all(v == 20 for v in arr)
    True

    # Sequence[float | nan | None], len(value) >= 4
    >>> arr = _to_float_array([20, None, float("nan"), 10, 0, 0], 4, -5.0)  # 6-len input
    >>> tuple(arr) == (20, -5.0, -5.0, 10)
    True

    # Sequence[float | float('nan') | None], len(value) < 4
    >>> arr = _to_float_array([20, None], 4, -5.0)
    >>> tuple(arr) == (20, -5.0, -5.0, -5.0)
    True

    """
    if value is None or _is_nan(value):
        return array('f', (default,) * length)
    if isinstance(value, (float, int)):
        return array('f', (value,) * length)

    arr = array('f', (default,) * length)
    try:
        for i in range(length):
            arr[i] = _to_value(value[i], default)
    except IndexError:
        pass
    return arr




# [ MINOR COMPONENTS ]  (major components/helpers of minor ones)

class _RectGetter(Protocol):
    """A function that returns an item's size and position; as
    `(width, height, x_pos, y_pos, is_visible)`.
    """
    def __call__(self, item: Item) -> tuple[int, int, int, int, bool]: ...


def _get_item_rect(item: Item) -> tuple[int, int, int, int, bool]:
    d  = _item_get_state(item)
    if 'visible' in d:
        return *d['rect_size'], *d['pos'], d['visible']  # type: ignore
    return *d['rect_size'], *d['pos'], _item_get_config(item)['show']  # type: ignore




class _RectSetter(Protocol):
    """A function that updates an item's size and position via
    `dearpygui.configure_item`.
    """
    def __call__(self, item: Item, x_pos: int, y_pos: int, width: int, height: int, show: bool) -> Any: ...


def _set_item_rect(item: Item, x_pos: int, y_pos: int, width: int, height: int, show: bool):
    _item_set_config(
        item,
        width=int(width),
        height=int(height),
        pos=(int(x_pos), int(y_pos)),
        show=show,
    )

def _set_text_rect(item: Item, x_pos: int, y_pos: int, width: int, height: int, show: bool):
    _item_set_config(
        item,
        pos=(int(x_pos), int(y_pos)),
        show=show,
    )




class _Positioner(Protocol):
    """Calculates the final position of an item in the occupying cell
    or cellspan.
    """
    def __call__(self, item_wt: float, item_ht: float, cell_x: float, cell_y: float, cell_wt: float, cell_ht: float, /) -> tuple[float, float]: ...


_ANCHOR_MAP: dict[str, _Positioner] = {}

def _get_positioner(key: str):
    try:
        return _ANCHOR_MAP[key.casefold()]
    except AttributeError:
        return _ANCHOR_MAP[key]  # raise `KeyError`

def _anchor_position(*anchors: str) -> Callable[[_Positioner], _Positioner]:
    def register_anchor_fn(fn: _Positioner) -> _Positioner:
        for s in anchors:
            _ANCHOR_MAP[s.lower()] = fn  # type: ignore
        return fn
    return register_anchor_fn


@_anchor_position("n", "north")
def _anchor_position_N(item_wt: float, item_ht: float, cell_x: float, cell_y: float, cell_wt: float, cell_ht: float) -> tuple[float, float]:
    return (cell_wt - item_wt) / 2 + cell_x, cell_y  # center x

@_anchor_position("ne", "northeast")
def _anchor_position_NE(item_wt: float, item_ht: float, cell_x: float, cell_y: float, cell_wt: float, cell_ht: float) -> tuple[float, float]:
    return (cell_wt - item_wt) + cell_x, cell_y

@_anchor_position("e", "east")
def _anchor_position_E(item_wt: float, item_ht: float, cell_x: float, cell_y: float, cell_wt: float, cell_ht: float) -> tuple[float, float]:
    return (cell_wt - item_wt) + cell_x, (cell_ht - item_ht) / 2 + cell_y  # center y

@_anchor_position("se", "southeast")
def _anchor_position_SE(item_wt: float, item_ht: float, cell_x: float, cell_y: float, cell_wt: float, cell_ht: float) -> tuple[float, float]:
    return (cell_wt - item_wt) + cell_x, (cell_ht - item_ht) + cell_y

@_anchor_position("s", "south")
def _anchor_position_S(item_wt: float, item_ht: float, cell_x: float, cell_y: float, cell_wt: float, cell_ht: float) -> tuple[float, float]:
    return (cell_wt - item_wt) / 2 + cell_x, (cell_ht - item_ht) + cell_y  # center x

@_anchor_position("sw", "southwest")
def _anchor_position_SW(item_wt: float, item_ht: float, cell_x: float, cell_y: float, cell_wt: float, cell_ht: float) -> tuple[float, float]:
    return cell_x, (cell_ht - item_ht) + cell_y

@_anchor_position("w", "west")
def _anchor_position_W(item_wt: float, item_ht: float, cell_x: float, cell_y: float, cell_wt: float, cell_ht: float) -> tuple[float, float]:
    return cell_x, (cell_ht - item_ht) / 2 + cell_y  # center y

@_anchor_position("nw", "northwest")
def _anchor_position_NW(item_wt: float, item_ht: float, cell_x: float, cell_y: float, cell_wt: float, cell_ht: float) -> tuple[float, float]:
    return cell_x, cell_y

@_anchor_position("c", "center", "centered")
def _anchor_position_C(item_wt: float, item_ht: float, cell_x: float, cell_y: float, cell_wt: float, cell_ht: float) -> tuple[float, float]:
    return (cell_wt - item_wt) / 2 + cell_x, (cell_ht - item_ht) / 2 + cell_y  # center x & y




@dataclasses.dataclass(slots=True)
class ItemData:
    """Contains the size and placement information of an item attached
    to a `Grid` object.

    Attributes:
        * item (int | str): The target item's integer uuid or string alias.

        * cellspan (tuple[int, int, int, int]): Contains coordinates for
        the target cell range, as `(col_start, row_start, col_end, row_end)`.

        * max_size (tuple[int, int]): Indicates the item's maximum width
        and height (min 0 per value).

        * padding (tuple[float, float, float, float]): Top-level padding
        override values.

        * positioner (_Positioner): Function used to calculate the item's
        final position.

        * rect_setter (_RectSetter): Callable used to update the item's size,
        position, and visibility status.

        * is_text (bool): If True, the target item an `mvText` object.
    """
    item       : Item
    cellspan   : tuple[int, int, int, int]
    max_size   : Sequence[float]
    padding    : Sequence[float]
    positioner : _Positioner
    rect_setter: _RectSetter
    is_text    : bool

    def __hash__(self):
        return hash(self.item)





# [ MAJOR COMPONENTS ]

class _GridSetting(Generic[_T]):
    __slots__ = ("_key",)

    def __init__(self, key: str = '', /):
        self._key = key

    def __set_name__(self, cls: type, name: str):
        self._key = self._key or name

    @overload
    def __get__(self, inst: Any, cls: type | None = ...) -> _T: ...
    @overload
    def __get__(self, inst: None, cls: type | None = None) -> Self: ...
    def __get__(self, inst: Any, cls: Any = None):
        if inst is None:
            return self
        return inst.configuration()[self._key]

    def __set__(self, inst: Any, value: Any):
        inst.configure(**{self._key: value})


@dataclasses.dataclass(init=False)
class _GridComponent:
    __slots__ = ('_label', '_spacing', '_padding')

    label  : _GridSetting[str]            = dataclasses.field(default=_GridSetting('label'))
    spacing: _GridSetting['array[float]'] = dataclasses.field(default=_GridSetting('spacing'))
    padding: _GridSetting['array[float]'] = dataclasses.field(default=_GridSetting('padding'))

    def __init__(self, *, label: str | None = None, spacing: FloatV = None, padding: Vec2 | FloatV = None, **kwargs):
        self.configure(label=label, spacing=spacing, padding=padding, **kwargs)

    def configure(self, **kwargs) -> None:
        """Update the object's high-level settings.

        Accepts all arguments accepted by the object's constructor
        as keyword arguments.
        """
        if 'label' in kwargs:
            self._label = kwargs['label'] or ''
        if 'padding' in kwargs:
            self._padding = _to_float_arr(kwargs['padding'], 2, NaN)
        if 'spacing' in kwargs:
            spacing = _to_value(kwargs['spacing'], None)
            if spacing is None:
                spacing = NaN
            else:
                spacing = max(0.0, spacing)
            self._spacing = spacing

    def configuration(self) -> dict[str, Any]:
        """Return a mapping of the object's settings and their current values."""
        return {f:getattr(self, f'_{f}') for f in self.__dataclass_fields__}


@dataclasses.dataclass(init=False)
class Slot(_GridComponent):
    """A row or column in a table-like structure.

    Attributes:
        * label: Informal name for the slot.

        * spacing (float): Used to offset the slot's position during a draw
        event. Is ignored for calculations when NaN.

        * padding (array[float, float]): Used as the left/upper and right/lower
        (for columns and rows respectively) padding values for cells in the slot
        during a draw event. Inner NaN values are ignored for calculations.

        * weight (float): During a draw event and when the slot's policy is
        evaluated as "sized", this affects how much the slot is scaled
        in proportion to the size of the grid and the total weight of all
        other slots in the axis.

        * size (int): The width/height (for columns and rows respectively) of
        the slot. When this value is not 0, a "fixed" sizing policy is enforced
        upon the slot, disabling dynamic resizing.


    A slot's sizing policy is not set, but determined during a draw event
    based on the slot's state. A "fixed" policy is enforced whenever a slot's
    `.size` attribute is greater than 0, and uses a "sized" policy otherwise.
    When "fixed", a slot is ensured to always be drawn to it's `.size`.

    A slot is unaware of its' role in the grid (column vs row). A higher-level
    object may indicate this through its' `.label` attribute.
    """
    __slots__ = ('_state', '_size', '_weight')

    weight: _GridSetting[float] = dataclasses.field(default=_GridSetting('weight'))
    size  : _GridSetting[int]   = dataclasses.field(default=_GridSetting('size'))

    @overload
    def __init__(self, *, label: str = ..., spacing: FloatV = ..., padding: Vec2 | FloatV = ..., weight: FloatV = ..., size: FloatV = ...) -> None: ...  # type: ignore
    def __init__(self, *, weight: Any = 1.0, size: Any = 0, **kwargs) -> None:
        """Args:
            * label: Used to update `self.label`. None is treated as ''.

            * spacing: Used to update `self.spacing`. A minimum 0 is used
            for non-NaN values. None is treated as NaN.

            * padding: Used to update `self.padding`. When the value is a
            sequence, the first two items of the sequence are used as the
            left/upper and right/lower (for columns and rows respectively)
            padding values, where inner None values are treated as NaN.
            Otherwise, the value is applied to all padding values while
            treating NaN and None as `(NaN, NaN)`.

            * weight: Used to update `self.weight`. A minimum of 0 is used
            for non-NaN values. None is treated as NaN.

            * size: Used to update `self.size`. A minimum of 0 is used for
            non-NaN values. None is treated as NaN.
        """
        # managed by the parenting `Grid` during a draw event
        self._state = cast(
            dict[Literal['pos' ,'size', 'spacing', 'padding'], Any],
            {
                'pos'    : 0,
                'size'   : 0,
                'spacing': 0,
                'padding': (0, 0),
            }
        )
        super().__init__(weight=weight, size=size, **kwargs)

    @overload
    def configure(self, *, label: str = ..., spacing: FloatV = ..., padding: Vec2 | FloatV = ..., weight: FloatV = ..., size: FloatV = ...) -> None: ... # type: ignore
    def configure(self, **kwargs) -> None:
        if 'weight' in kwargs:
            self._weight = max(0.0, _to_value(kwargs['weight'], 0.0))
        if 'size' in kwargs:
            self._size = max(0, int(_to_value(kwargs['size'], 0)))
        super().configure(**kwargs)

    def configuration(self) -> dict[Literal['label', 'spacing', 'padding', 'weight', 'size'], Any]: ...
    del configuration


@dataclasses.dataclass(init=False)
class Axis(_GridComponent, Iterable[Slot], Sized):
    """A collection of rows or columns.

    Attributes:
        * label (str): Informal name for the axis.

        * spacing (float): Used as a fallback value to offset the slot's
        position during a draw event. Is ignored for calculations when NaN.

        * padding (array[float, float]): Fallback values used during a draw
        event when a slot contained in the axis does not specify a padding
        value. Inner NaN values are ignored for calculations.

        * length (int): Evaluates to `len(self)`. Setting a value resizes
        the length of the axis; adding or removing slots as necessary.


    Instances are treated as immutable, unsigned pseudo-integrals
    for most operations; the value of which is a representation
    of the number of slots in the axis. However, in-place operations
    directly affect the number of slots contained.

    Axes objects behave similar to lists. Slots are identified by their
    positions (index) in the axis. Indexing the axis returns the `Slot`
    object for that row or column.

    When the axis' number of slots is changed, slots are always appended
    to, or are removed from, the end of the axis.

    Operations that update the number of slots in the axis are
    thread-safe.

    An axis object is unaware of its' role in the grid i.e. x
    (columns) vs y (rows). A higher-level object may indicate this
    through its' `.label` attribute.
    """
    __slots__ = ('_slots', '_lock')

    length: _GridSetting[int] = dataclasses.field(default=_GridSetting('length'))

    @overload
    def __init__(self, length: int = ..., *, label: str = ..., spacing: FloatV = ..., padding: Vec2 | FloatV = ...) -> None: ...  # type: ignore
    def __init__(self, length: int = 0, **kwargs) -> None:
        """Args:
            * length: Integer value used to update `self.length`.

            * label: Used to update `self.label`. None is treated as ''.

            * spacing: Used to update `self.spacing`. A minimum 0 is used
            for non-NaN values. None is treated as NaN.

            * padding: Used to update `self.padding`. When the value is a
            sequence, the first two items of the sequence are used as the
            left/upper and right/lower (for columns and rows respectively)
            padding values, where inner None values are treated as NaN.
            Otherwise, the value is applied to all padding values while
            treating NaN and None as `(NaN, NaN)`.
        """
        self._slots = cast(list[Slot], [])
        self._lock  = threading.Lock()
        super().__init__(length=length, **kwargs)

    @overload
    def configure(self, *, label: str = ..., spacing: FloatV = ..., padding: Vec2 | FloatV = ..., length: int = ...): ...  # type: ignore
    def configure(self, **kwargs):
        if 'length' in kwargs:
            self.resize(kwargs['length'])
        super().configure(**kwargs)

    def configuration(self) -> dict[Literal['label', 'spacing', 'padding', 'length'], Any]: ...
    del configuration

    def __str__(self):
        return str(self._slots)

    def __bool__(self):
        return bool(self._slots)

    def __len__(self) -> int:
        # TODO: check for concurrency issues
        return len(self._slots)

    __int__ = __index__ = __len__

    def __float__(self):
        return float(len(self))

    def __complex__(self):
        return complex(len(self))

    def __lt__(self, other: Any) -> bool:
        return len(self) < other

    def __le__(self, other: Any) -> bool:
        return len(self) <= other

    def __eq__(self, other: Any) -> bool:  # XXX: `__hash__ == None`
        return len(self) == other

    def __ne__(self, other: Any) -> bool:
        return len(self) != other

    def __gt__(self, other: Any) -> bool:
        return len(self) > other

    def __ge__(self, other: Any) -> bool:
        return len(self) >= other

    def slot_weight(self) -> float:
        """Return the weight sum of all "sized" slots in the axis."""
        with self._lock:
            return sum(s._weight for s in self._slots if not s._size)

    def slot_size(self) -> int:
        """Return the size sum of all "fixed" slots in the axis."""
        with self._lock:
            return sum(s._size for s in self._slots)


    # Number Behaviors (no bitwise)

    def __pos__(self):
        return +len(self)

    def __neg__(self):
        return -len(self)

    def __add__(self, x: _N) -> _N:
        return len(self) + x

    def __sub__(self, x: _N) -> _N:
        return len(self) - x

    def __mul__(self, x: _N) -> _N:
        return len(self) * x

    def __truediv__(self, x: _N) -> _N:
        return len(self) * x

    def __floordiv__(self, x: _N) -> _N | float:
        return len(self) // x

    def __mod__(self, x: _N) -> _N:
        return len(self) % x

    def __divmod__(self, x: Any) -> tuple[float, float]:
        length = len(self)
        return (length // x, length % x)

    def __pow__(self, x: int, mod: int | None = None) -> int:
        return pow(len(self), x, mod)

    def __round__(self, ndigits: int = 0):
        return round(len(self), ndigits)

    __trunc__ = __floor__ = __ceil__ = __abs__ = __int__


    # Sequence Behaviors/Methods

    def __getitem__(self, index: SupportsIndex) -> Slot:
        return self._slots[index]

    def __iter__(self) -> Iterator[Slot]:
        yield from self._slots

    def __iadd__(self, x: int) -> Self:
        """
        >>> x = Axis(4)
        >>> x += 8
        >>> len(x)
        12
        >>> x += 0
        >>> len(x)
        12
        >>> x += -10  # __isub__
        >>> len(x)
        2
        """
        if not x:
            return self
        if x > 0:
            with self._lock:
                self._slots.extend(Slot() for _ in range(x))
                return self
        return self.__isub__(abs(x))

    def __isub__(self, x: int) -> Self:
        """
        >>> x = Axis(12)
        >>> x -= 8
        >>> len(x)
        4
        >>> x -= 0
        >>> len(x)
        4
        >>> x -= -10  # __iadd__
        >>> len(x)
        12
        """
        if not x:
            return self
        if x > 0:
            with self._lock:
                del self._slots[-x:]
                return self
        return self.__iadd__(abs(x))

    def resize(self, length: int):
        """Add/remove slots from the axis so that it contains the number
        of slots specified.

        Awaits internal lock release.

        Args:
            * length: A positive number indicating the target length of
            the axis.


        >>> x = Axis(8)
        >>> x.resize(2)  # trim
        >>> len(x)
        2
        >>> x.resize(12)  # extend
        >>> len(x)
        12
        """
        if length < 0:
            raise ValueError(f'`length` cannot be less than zero (got {length!r}).')
        return self.__iadd__(length-len(self))

    def __imul__(self, x: int):  # NOTE: ROUNDS DOWN
        """
        >>> x = Axis(4)
        >>> x *= 4
        >>> len(x)
        16
        >>> x *= 0.73  # trunc 11.68
        >>> len(x)
        11
        """
        if x < 0:
            raise ValueError(f'cannot multiply slots by a negative number (got {x!r}).')
        return self.resize(int(len(self) * x))

    def __itruediv__(self, x: float):  # NOTE: ROUNDS DOWN (floor division)
        """
        >>> x = Axis(16)
        >>> x /= 2
        >>> len(x)
        8
        >>> x /= 3  # floor div
        2
        """
        if x < 0:
            raise ValueError(f"cannot divide slots by a negative number (got {x!r}).")
        return self.resize(len(self) // x)  # type: ignore

    __ifloordiv__ = __itruediv__

    def insert(self, index: SupportsIndex):
        """Adds a new row/column at the specified index.

        Awaits internal lock release.

        Args:
            * index: Position of the new slot.
        """
        with self._lock:
            self._slots.insert(index, Slot())

    def remove(self, index: SupportsIndex = -1):
        """Delete the row/column at the specified index, or the last
        row/column if not specified.

        Awaits internal lock release.

        Args:
            * index: Target slot index.
        """
        with self._lock:
            self._slots.pop(index)


@dataclasses.dataclass(init=False)
class Grid(_GridComponent):
    """Layout manager for Dear PyGui, emulating a table-like structure.

    Attributes:
        * cols (Axis): Living representation of the grid's columns (x-axis).
        See the `Axis` class for more information.

        * rows (Axis): Living representation of the grid's rows (y-axis).
        See the `Axis` class for more information.

        * target (int | str): The integer identifier or string alias of the
        item used as the grid's positional reference. Additionally, it's
        visibility state is used to determine whether or not to display (draw)
        the grid. The size of the target item is also used as fallback values
        as necessary when `self.width` and/or `self.height` are not applicable.

        * label (str): An informal name for the grid.

        * spacing (array[float, float]): Values used to offset a slot's (columns
        and rows, respectively) position during a draw event when both
        `slot.spacing` and `axis.spacing` are not applicable (min. 0 per value).

        * padding (array[float, float, float, float]): Used as the left, upper,
        right, and lower padding values during a draw event when item-level
        padding override values were not provided and both `slot.spacing` and
        `axis.spacing` are not applicable. The first and third values are used
        for column padding, while the second and fourth values are used for rows
        (min. 0 per value).

        * width (int): The horizontal size of the grid. Can be used to extend
        the grid beyond the default view of the target item, allowing for items
        to be positioned in their parent's scroll area. If 0, the width of the
        target item is used instead (min. 0).

        * height (int): The vertical size of the grid. Can be used to extend
        the grid beyond the default view of the target item, allowing for items
        to be positioned in their parent's scroll area. If 0, the height of the
        target item is used instead (min. 0).

        * offsets (array[float, float, float, float]): Similar to padding, but
        targets the left, upper, right, and lower edges of the grid itself and
        not its' slots. Can be used to expand or shrink the size of the grid's
        content region (min. 0 per value).

        * rect_getter (_RectGetter): Called before attempting to draw the grid
        to fetch the target item's size, position, and visibility status. The
        default implementation is a light wrapper around
        `dearpygui.get_item_state`. It may also call
        `dearpygui.get_item_configuration` if one or more details cannot be
        accessed by reading the item's state.

        * overlay (bool): If True, a visual representation of the grid will be
        displayed over it.

        * show (bool): If True, the grid and its' content will be visible.


    The grid calculates the positions, sizes, and visibility states of
    itself and attached items during a draw event; triggered by calling the
    grid. The grid must be frequently drawn in order for it to operate as
    expected, performing best when registered as the resize callback for its'
    target item or the parenting root-level window. However, the grid can
    registered as any Dear PyGui callback or even called directly.

    The grid does not use or create items to assist with the layout. Any item
    generated by the grid is used for its' overlay, enabled when `self.overlay`
    is True. A class-level `mvViewportDrawlist` item is created when the first
    instance of `Grid` is created, while a `mvDrawLayer` item is created per
    `Grid` instance. When `self.overlay` is False, no additional items are
    created.

    When a draw event occurs, the "draw" step is skipped when the grid is
    not visible. Additionally, the grid, its content, and overlay are drawn
    and displayed only if the visibility value returned from `self.rect_getter`
    is True, regardless of the value of `self.show` and `self.overlay`.

    Items can be attached to the grid via the `.push` method. When sizing and
    positioning an item during a draw event, settings provided when attatching
    the item (padding, spacing, etc.) are used over slot, axis and grid settings.
    If setting(s) are unspecified (usually indicated by None or NaN), the grid
    falls back to using settings found on the slot, then axis, etc.

    There are situations where an item may be hidden when drawing
    the grid:
        * the item's cellspan is outside of the grid's current
        range
        * given the constraints of the slots in the target cellspan,
        the size of the item is calculated to be less than 1 pixel

    Any item attached to the grid that does not exist at the time
    of a draw event is silently dropped from the grid, and will not
    be managed further.

    For more information, see `Grid.__init__`, `Grid.__call__` and `Grid.push`.

    """
    __slots__ = (
        # internal
        '_lock',
        '_item_data',
        '_trashbin',
        '_drawlayer',
        # configuration
        'rows',
        'cols',
        '_width',
        '_height',
        '_offsets',
        '_target',
        '_rect_getter',
        '_overlay',
        '_show',
        # other
        '__weakref__',
    )

    target     : _GridSetting[Item]           = dataclasses.field(default=_GridSetting("target"))
    spacing    : _GridSetting['array[float]'] = dataclasses.field(default=_GridSetting("spacing"))  # float -> array[float]
    width      : _GridSetting[int]            = dataclasses.field(default=_GridSetting("width"))
    height     : _GridSetting[int]            = dataclasses.field(default=_GridSetting("height"))
    offsets    : _GridSetting['array[float]'] = dataclasses.field(default=_GridSetting("offsets"))
    rect_getter: _GridSetting[_RectGetter]    = dataclasses.field(default=_GridSetting("rect_getter"))
    overlay    : _GridSetting[bool]           = dataclasses.field(default=_GridSetting("overlay"))
    show       : _GridSetting[bool]           = dataclasses.field(default=_GridSetting("show"))

    __drawlist = '[mvViewportDrawlist] Grid'

    @overload
    def __init__(self, cols: int = ..., rows: int = ..., target: Item = ..., *, label: str | None = ..., padding: Vec4 | FloatV = ..., spacing: Vec2 | FloatV = ..., width: FloatV = ..., height: FloatV = ..., offsets: Vec4 | FloatV = ..., rect_getter: _RectGetter = ..., overlay: bool = ..., show: bool = ...) -> None: ...  # type: ignore
    def __init__(
        self,
        cols       : int  = 1,
        rows       : int  = 1,
        target     : Item = 0,
        *,
        width      : Any  = None,
        height     : Any  = None,
        offsets    : Any  = None,
        rect_getter: Any  = None,
        overlay    : bool = False,
        show       : bool = True,
        **kwargs,
    ) -> None:
        """Args:
            * cols: Integer value used to update `self.cols` via `self.cols.resize`.
            Sets the number of slots in the x-axis.

            * rows: Integer value used to update `self.rows` via `self.rows.resize`.
            Sets the number of slots in the y-axis.

            * target: Integer or string value used to update `self.target`. Non-
            optional in most cases.

            * label: Used to update `self.label`. None is treated as ''.

            * rect_getter: Used to update `self.rect_getter`. Should be None, or a
            callable accepting `self.target` as a positional argument and returns a
            5-tuple containing a related width, height, x-position, y-position, and
            visibility status. None is treated as the internal default value.

            * width: Used to update `self.width`. Sets a minimum value of 0.
            None is treated as 0.

            * height: Used to update `self.height`. Sets a minimum value of 0.
            None is treated as 0.

            * offsets: Used to update `self.offsets`. When the value is a
            sequence, the first four items are used as the grid's left, upper,
            right, and lower inner padding values, where inner None values are
            treated as 0. Otherwise, the value is applied to all padding values
            while treating NaN and None as `(0, 0, 0, 0)`.

            * padding: Used to update `self.padding`. When the value is a
            sequence, the first four items are used as the left, upper,
            right, and lower padding values for slots, where inner None values
            are treated as 0. Otherwise, the value is applied to all padding values
            while treating NaN and None as `(0, 0, 0, 0)`.

            * spacing: Used to update `self.spacing`. When the value is a
            sequence, the first two items are used as column and row spacing values,
            where inner None values are treated as 0. Otherwise, the value is applied
            to all spacing values while treating NaN and None as `(0, 0)`.

            * overlay: Used to update `self.overlay`. Evaluated as a boolean, True
            will cause the overlay to be displayed once a draw event has been
            triggered.

            * show: Used to update `self.show`. Evaluated as a boolean,
            False will hide the grid, its' overlay, and all attached items (and
            vice-versa).
        """
        _create_context()
        if not dearpygui.does_item_exist(self.__drawlist):
            type(self).__drawlist = dearpygui.add_viewport_drawlist()
        self._drawlayer = dearpygui.add_draw_layer(parent=self.__drawlist)
        self._item_data = cast(set[ItemData], set())
        self._trashbin  = set()
        self._lock      = threading.Lock() if not sys.gettrace() else threading.RLock()

        self.cols = Axis(0, label='x')
        self.rows = Axis(0, label='y')

        super().__init__(
            rows=rows,
            cols=cols,
            width=width,
            height=height,
            offsets=offsets,
            target=target,
            rect_getter=rect_getter,
            overlay=overlay,
            show=show,
            **kwargs,
        )

    def _clean_cache(self):
        # currently called before or after:
        #   * updating the grid's settings
        #   * updating `self._item_data`
        #   * drawing the grid
        self._item_data.difference_update(self._trashbin)
        self._trashbin.clear()

    @overload
    def configure(self, *, cols: int = ..., rows: int = ..., target: Item = ..., label: str | None = ..., padding: Vec4 | FloatV = ..., spacing: Vec2 | FloatV = ..., offsets: Vec4 | FloatV = ..., width: FloatV = ..., height: FloatV = ..., rect_getter: _RectGetter = ..., overlay: bool = ..., show: bool = ...): ...  # type: ignore
    def configure(self, **kwargs):
        """Update the grid's settings.

        Awaits internal lock release.
        """
        with self._lock:
            # slots
            if 'cols' in kwargs:
                cols = kwargs['cols']
                if not isinstance(cols, int):
                    raise TypeError(f'expected int for `cols` (got {type(cols)!r}).')
                self.cols.resize(kwargs['cols'])
            if 'rows' in kwargs:
                rows = kwargs['rows']
                if not isinstance(rows, int):
                    raise TypeError(f'expected int for `rows` (got {type(rows)!r}).')
                self.rows.resize(kwargs['rows'])

            if 'target' in kwargs or 'rect_getter' in kwargs:
                target      = kwargs.get('target', getattr(self, '_target', 0)) or 0
                rect_getter = kwargs.get('rect_getter', getattr(self, '_rect_getter', _get_item_rect)) or _get_item_rect
                if not target and rect_getter is _get_item_rect:
                    raise ValueError(f'`target` required when using the default rect-getter.')
                try:
                    _, _, _, _, _ = rect_getter(target)
                except ValueError:
                    raise TypeError(f'expected `rect_getter` to return a 5-tuple, got {type(rect_getter)!r}.')
                except TypeError:
                    if not callable(rect_getter):
                        raise TypeError(f'expected callable for `rect_getter`, got {type(rect_getter)!r}.') from None
                    raise
                except SystemError:
                    if target and dearpygui.does_item_exist(target):
                        raise

                self._target      = target
                self._rect_getter = rect_getter

            # sizing
            if 'width' in kwargs:
                self._width = int(max(0, _to_value(kwargs['width'], 0)))
            if 'height' in kwargs:
                self._height = int(max(0, _to_value(kwargs['height'], 0)))

            # offsets
            if 'offsets' in kwargs:
                self._offsets = _to_float_arr(kwargs['offsets'], 4, 0.0)
            if 'padding' in kwargs:
                self._padding = _to_float_arr(kwargs.pop('padding'), 4, 0.0)  # override
            if 'spacing' in kwargs:
                self._spacing = _to_float_arr(kwargs.pop('spacing'), 2, 0.0)  # override

            # visibility
            if 'overlay' in kwargs:
                overlay = bool(kwargs['overlay'])
                _item_set_config(self._drawlayer, show=overlay)
                self._overlay = overlay
            if 'show' in kwargs:
                show = bool(kwargs['show'])
                if not show:
                    _item_set_config(self._drawlayer, show=False)
                for item_data in self._item_data:
                    try:
                        _item_set_config(item_data.item, show=show)
                    except SystemError:
                        if not dearpygui.does_item_exist(item_data.item):
                            self._trashbin.add(item_data)
                        else:
                            raise
                self._show = show

            super().configure(**kwargs)
            self._clean_cache()

    def configuration(self):
        """Return the grid's various settings and values."""
        cfg = {'cols': len(self.cols), 'rows': len(self.rows)}
        cfg.update({f:getattr(self, f'_{f}') for f in self.__dataclass_fields__})
        return cfg

    def _pop(self, item: Item):
        # Two instances of `ItemData` can potentially exist for the same item
        # since refs can be integers or strings. Try to remove all of them just
        # in case.
        if not dearpygui.does_item_exist(item):
            self._trashbin.add(item)
        else:
            if isinstance(item, str):
                refs = (item, dearpygui.get_alias_id(item), dearpygui.get_item_alias(item))
            else:
                refs = (item, dearpygui.get_item_alias(item))
            self._trashbin.update(refs)
            _item_set_config(item, pos=())
        self._clean_cache()

    def pop(self, item: Item):
        """Release an item from the grid.

        Awaits internal lock release.
        """
        with self._lock:
            self._pop(item)

    def clear(self):
        """Release all items from the grid.

        Awaits internal lock release.
        """
        with self._lock:
            for item_data in self._item_data:
                try:
                    _item_set_config(item_data.item, pos=())
                except SystemError:
                    pass
            self._trashbin.update(self._item_data)
            self._clean_cache()

    ANCHORS = tuple(s.lower() for s in _ANCHOR_MAP)

    @overload
    def push(self, item: Item, col: int, row: int, /, *, anchor: str = ..., rect_setter: _RectSetter = ..., max_size: Vec2 | FloatV = ..., padding: Vec4 | FloatV = ...): ...
    @overload
    def push(self, item: Item, col: int, row: int, /, *, anchor: str = ..., rect_setter: _RectSetter = ..., max_width: int = ..., max_height: int = ..., padding: Vec4 | FloatV = ...): ...
    @overload
    def push(self, item: Item, col: int, row: int, /, *, anchor: str = ..., rect_setter: _RectSetter = ..., max_size: Vec2 | FloatV = ..., x1_pad: int = ..., y1_pad: int = ..., x2_pad: int = ..., y2_pad: int = ...): ...
    @overload
    def push(self, item: Item, col: int, row: int, /, *, anchor: str = ..., rect_setter: _RectSetter = ..., max_width: int = ..., max_height: int = ..., x1_pad: int = ..., y1_pad: int = ..., x2_pad: int = ..., y2_pad: int = ...): ...
    @overload
    def push(self, item: Item, cell_start: tuple[int, int] | Sequence[int], cell_stop: tuple[int, int] | Sequence[int] | None = ..., /, *, anchor: str = ..., rect_setter: _RectSetter = ..., max_size: Vec2 | FloatV = ..., padding: Vec4 | FloatV = ...): ...
    @overload
    def push(self, item: Item, cell_start: tuple[int, int] | Sequence[int], cell_stop: tuple[int, int] | Sequence[int] | None = ..., /, *, anchor: str = ..., rect_setter: _RectSetter = ..., max_width: int = ..., max_height: int = ..., padding: Vec4 | FloatV = ...): ...
    @overload
    def push(self, item: Item, cell_start: tuple[int, int] | Sequence[int], cell_stop: tuple[int, int] | Sequence[int] | None = ..., /, *, anchor: str = ..., rect_setter: _RectSetter = ..., max_size: Vec2 | FloatV = ..., x1_pad: int = ..., y1_pad: int = ..., x2_pad: int = ..., y2_pad: int = ...): ...
    @overload
    def push(self, item: Item, cell_start: tuple[int, int] | Sequence[int], cell_stop: tuple[int, int] | Sequence[int] | None = ..., /, *, anchor: str = ..., rect_setter: _RectSetter = ..., max_width: int = ..., max_height: int = ..., x1_pad: int = ..., y1_pad: int = ..., x2_pad: int = ..., y2_pad: int = ...): ...
    def push(self, item: Item, cell_start: Any, cell_stop: Any = None, /, *, anchor: str = 'c', rect_setter: _RectSetter = _set_item_rect, max_size: Any = None, max_width: Any = None, max_height: Any = None, padding: Any = None, x1_pad: Any = None, y1_pad: Any = None, x2_pad: Any = None, y2_pad: Any = None):
        """Attach an item to the grid.

        Awaits internal lock release.

        Args:
            * item: Integer uuid or string alias of an item.

            * col, cell_start: Index of the cell's column that will contain the
            item, OR a 2-item sequence containing both the column and row indices
            of the initial cell of the cell range the item will occupy.

            * row, cell_stop: Index of the cell's row that will contain the item,
            OR a 2-item sequence containing both the column and row indices of the
            last cell of the cell range the item will occupy.

            * anchor: A (inter)cardinal direction name (e.g. "north", "southwest")
            or abbreviation (e.g. "n", "sw") indicating the area of the cell the
            item will "stick" to. "center", "centered", or "c" will result in the
            item being centered in the specified cell or cellspan (default). Value
            is case-insensitive.

            * rect_setter: A callable accepting *item*, x-position, y-position,
            width, height, and visible state as positional arguments that updates
            the item's size, position, and visibility. By default, it is a
            simple wrapper around `dearpygui.configure_item`.

            * max_size: The maximum width and height the item will scale to. Inner
            NaN-like values are not applied.

            * max_width: Individual width component (first value) of *max_size*.
            Ignored when *max_size* is specified.

            * max_height: Individual height component (second value) of *max_size*.
            Ignored when *max_size* is specified.

            * padding: A 4-item sequence whose values will be used to override the
            left, upper, right, and/or lower-padding values of the first/last slots
            in the target cell range, OR, a non-sequence value that will be used
            as the override value for all padding components. Inner NaN-like values
            are not applied as an overriding value.

            * x1_pad: Overrides the first `.padding` value of the first column
            (left-padding) in the target cell range. Ignored when *padding* is
            specified.

            * y1_pad: Overrides the first `.padding` value of the first row
            (upper-padding) in the target cell range. Ignored when *padding* is
            specified.

            * x2_pad: Overrides the first `.padding` value of the last column
            (right-padding) in the target cell range. Ignored when *padding* is
            specified.

            * y2_pad: Overrides the first `.padding` value of the last row
            (lower-padding) in the target cell range. Ignored when *padding* is
            specified.


        An item attached to the grid must support sizing via `width` and `height`
        configurations, positioning via `pos`, and visibility updates via `show`.
        Otherwise, consider attaching a compatible parenting, container item
        instead. Alternatively, a custom *rect_setter* can be passed to enforce
        item compatibility.

        When it is desired for the item to occupy a single cell, the second and
        third positional arguments are the target cell's column and row indices,
        respectively. Alternatively, a 2-item sequence can be passed as the
        second positional argument containing both indices while leaving the
        third positional argument unspecified. To push an item onto a range of
        cells, the second and third positional arguments must both be 2-item
        sequences containing indices of the endpoint cells in the range. During
        a draw event, the cell whose indices are nearest to `(0, 0)` is
        considered the "starting" cell in the range for that event.

        Column/row indices can be positive or negative integers. The "indexing"
        behavior mirrors typical Python subscript behavior; cell coordinates
        `(-1, 0)` would specify the cell at the last column and first row.

        The exact cell(s) used are evaluated during draw events; deterministic
        of the grid's state at the time of the event and *not* when the item is
        attached. A side effect of this is that column and row indices can be
        outside the grid's range when items are attached. If the target cell or
        cellspan is out-of-range during a draw event, the item is simply hidden.
        Additionally, passing a row index of 4 is not the same as -1 even if the
        grid only has 5 rows at the time of attachment since the exact row is
        not determined until a draw event occurs.

        During a draw event, *rect_setter* is called to update the item using
        the finalized size, position, and visibility values. This means that items
        attached to the grid must support sizing via `width` and `height`
        configurations, positioning via `pos`, and visibility updates via `show`.
        If an item does not support one of more of the mentioned configuration
        options, consider parenting it within a more compatible item and attaching
        that item instead. Otherwise, a custom *rect_setter* can be used force
        compatibility of otherwise incompatible items by choosing which settings
        are updated in their call to `dearpygui.configure_item` or, alternatively,
        update a different item entirely.

        An item's size is usually dependant on the size of the row(s) and column(s)
        they occupy. *max_size* (OR individual *max_width/height*) values can be
        passed to clamp the item's size as the cell or cellspan expands. Setting a
        minimum size per item is not supported, since having the content of a
        cell(span) extend beyond its' borders is not typical of a grid system.
        When desired, it is recommended to set the `size` attribute of the entire
        row or column. This sets a "fixed" sizing policy on the slot, preventing it
        from being dynamically sized. If an item *must* have a unique minimum size,
        a custom *rect_setter* can be set that sends other width/height values to
        `dearpygui.configure_item`.

        > NOTE: A grid object that calls *rect_setter* will be holding its' internal
        mutex. Under normal operation, a *rect_setter* that calls a grid's methods or
        invokes behavior documented to "await internal lock release" will result in
        a deadlock.
        """
        if hasattr(cell_start, '__iter__'):  # possible cell range
            try:
                x1, y1 = cell_start
            except ValueError:
                raise ValueError(
                    f'expected a 2-item sequence of cell coordinates for `cell_start`, got {len(cell_start)!r}.'
                ) from None

            try:
                x2, y2 = cell_stop
            except TypeError:
                if cell_stop is None:  # single cell
                    x2 = x1
                    y2 = y1
                else:
                    raise ValueError(
                        f'expected a 2-item sequence of cell coordinates for `cell_stop`, got {cell_stop!r}.'
                    ) from None
        else:  # single cell
            x1 = x2 = cell_start

            try:
                y1 = y2 = int(cell_stop)
            except TypeError:
                if cell_stop is not None:
                    raise TypeError(
                        f'expected int for `row`, got {type(cell_stop)!r}.'
                    ) from None
                raise TypeError(
                    f"expected 3 positional arguments for '{self.push.__qualname__}()' got 2."
                ) from None
        cellspan = int(x1), int(y1), int(x2), int(y2),

        if max_size is _is_nanlike(max_size):
            max_size = (max_width, max_height)
        max_size = _to_float_arr(max_size, 2, 0.0)
        for i in range(len(max_size)):
            max_size[i] = max(0.0, max_size[i])

        if padding is _is_nanlike(padding):
            padding = (x1_pad, y1_pad, x2_pad, y2_pad)
        padding = _to_float_arr(padding, 4, NaN)

        with self._lock:
            self._pop(item)

            # unique handling per item type
            is_text = False
            try:
                item_type = dearpygui.get_item_info(item)['type']
            except SystemError:
                # XXX: The item may not exist yet. It is very rare that this is
                # purposeful, but it happens? Allowing this can create gachas.
                pass
            else:
                if item_type == 'mvAppItemType::mvText':
                    is_text = True
                    if rect_setter == _set_item_rect:
                        rect_setter = _set_text_rect
                elif item_type == 'mvAppItemType::mvTable':
                    # BUG: Table items accept a `pos` keyword, but DPG doesn't
                    # do anything with it.
                    raise ValueError("`mvTable` items cannot be explicitly positioned.")

            item_data = ItemData(
                item=item,
                cellspan=cellspan,
                max_size=max_size,
                padding=padding,
                positioner=_get_positioner(anchor),
                rect_setter=rect_setter,
                is_text=is_text,
            )
            self._item_data.add(item_data)
            return item_data

    def _upd_slot_states(self, axis: Axis, area_size: float, index_offset: Literal[0, 1]):  # XXX: performance-sensitive
        area_c1_pad  = self._offsets[0 + index_offset]
        area_c2_pad  = self._offsets[2 + index_offset]
        content_size = area_size - area_c1_pad - area_c2_pad

        _slot_c1_pad, _slot_c2_pad = axis._padding
        if _slot_c1_pad != _slot_c1_pad:  # is NaN?
            _slot_c1_pad = self._padding[0 + index_offset]
        if _slot_c2_pad != _slot_c2_pad:  # is NaN?
            _slot_c2_pad = self._padding[2 + index_offset]

        _slot_size_offset = axis._spacing
        if _slot_size_offset != _slot_size_offset:  # is NaN?
            _slot_size_offset = self._spacing[0 + index_offset]

        slot_pos_offset = _slot_size_offset / 2.0

        slot_weight_size = (
            max(0, (content_size - axis.slot_size())) / max(1, axis.slot_weight())
        )

        alloc_size = 0.0
        for slot in axis:
            slot_state = slot._state

            slot_size_offset = slot._spacing
            if slot_size_offset != slot_size_offset:  # is NaN?
                slot_size_offset = _slot_size_offset
            slot_state['spacing'] = slot_size_offset

            slot_state['pos'] = alloc_size + area_c1_pad + slot_pos_offset

            slot_c1_pad, slot_c2_pad = slot._padding
            if slot_c1_pad != slot_c1_pad:  # is NaN?
                slot_c1_pad = _slot_c1_pad
            if slot_c2_pad != slot_c2_pad:  # is NaN?
                slot_c2_pad = _slot_c2_pad
            slot_state['padding'] = slot_c1_pad, slot_c2_pad

            slot_size = slot._size or slot_weight_size * slot._weight  # FIXED or SIZED
            slot_state['size'] = slot_size - _slot_size_offset
            alloc_size += slot_size

    def _upd_item_states(self):  # XXX: performance-sensitive
        rows = self.rows._slots
        cols = self.cols._slots

        n_cols = len(cols)
        n_rows = len(rows)

        for item_data in self._item_data:
            x1, y1, x2, y2 = item_data.cellspan
            # convert negative indexes
            x1 %= n_cols
            y1 %= n_rows
            x2 %= n_cols
            y2 %= n_rows
            # fix inversed cellspan
            if y1 > y2:
                y1, y2 = y2, y1
            if x1 > x2:
                x1, x2 = x2, x1

            try:
                col1_state = cols[x1]._state
                row1_state = rows[y1]._state
                col2_state = cols[x2]._state
                row2_state = rows[y2]._state
            except KeyError:  # hide the item - cellspan is out-of-range
                item_x_pos  = 0
                item_y_pos  = 0
                item_width  = 0
                item_height = 0
                item_show   = False
            else:
                x1_pad, y1_pad, x2_pad, y2_pad = item_data.padding

                if x1_pad != x1_pad:  # is NaN?
                    x1_pad = col1_state['padding'][0]
                cell_x_pos = col1_state['pos'] + x1_pad

                if y1_pad != y1_pad:  # is NaN?
                    y1_pad = row1_state['padding'][0]
                cell_y_pos  = row1_state['pos'] + y1_pad

                if x2_pad != x2_pad:  # is NaN?
                    x2_pad = col2_state['padding'][1]
                cell_width = col2_state['pos'] + col2_state['size'] - cell_x_pos - x2_pad

                if y2_pad != y2_pad:  # is NaN?
                    y2_pad = row1_state['padding'][1]
                cell_height = row2_state['pos'] + row2_state['size'] - cell_y_pos - y2_pad

                # XXX: It's really common for users to want to align text items.
                # Unfortunately, they must be handled differently since their size
                # is determined by their value and font used.
                if item_data.is_text:
                    try:
                        item_width, item_height = _item_get_state(item_data.item)['rect_size']
                        # if item_width > cell_width or item_height > cell_height:
                        #     item_width = item_height = 0
                    except SystemError:  # does not exist - dealt with below
                        item_width = item_height = 0
                else:
                    item_width, item_height = item_data.max_size
                    if not item_width or item_width > cell_width:
                        item_width = cell_width
                    if not item_height or item_height > cell_height:
                        item_height = cell_height

                if item_width < 1 or item_height < 1:  # hide the item - size too small
                    item_x_pos = 0
                    item_y_pos = 0
                    item_show  = False
                else:
                    item_x_pos, item_y_pos = item_data.positioner(
                        item_width,
                        item_height,
                        cell_x_pos,
                        cell_y_pos,
                        cell_width,
                        cell_height,
                    )
                    item_show = True

            try:
                item_data.rect_setter(
                    item_data.item,
                    int(item_x_pos),
                    int(item_y_pos),
                    int(item_width),
                    int(item_height),
                    item_show,
                )
            except SystemError:
                if not dearpygui.does_item_exist(item_data.item):
                    self._trashbin.add(item_data.item)
                else:
                    raise

    def _draw_outline(self, x_min: int, y_min: int, x_max: int, y_max: int, line_color: Sequence[int], pad_color: Sequence[int]):
        layer = self._drawlayer
        x1_pad, y1_pad, x2_pad, y2_pad = self._offsets

        # padding
        if x1_pad:
            dearpygui.draw_rectangle(
                (x_min, y_max),
                (x_min + x1_pad, y_min + y1_pad),
                color=(0, 0, 0, 0),
                fill=pad_color,  # type: ignore
                parent=layer,
            )
        if y1_pad:
            dearpygui.draw_rectangle(
                (x_min, y_min),
                (x_max - x2_pad, y_min + y1_pad),
                color=(0, 0, 0, 0),
                fill=pad_color,  # type: ignore
                parent=layer,
            )
        if x2_pad:
            dearpygui.draw_rectangle(
                (x_max, y_min),
                (x_max - x2_pad, y_max - y2_pad),
                color=(0, 0, 0, 0),
                fill=pad_color,  # type: ignore
                parent=layer,
            )
        if y2_pad:
            dearpygui.draw_rectangle(
                (x_max, y_max),
                (x_min + x1_pad, y_max - y2_pad),
                color=(0, 0, 0, 0),
                fill=pad_color,  # type: ignore
                parent=layer,
            )

        # outline
        dearpygui.draw_rectangle(
            (x_min, y_min), (x_max, y_max),
            fill=(0, 0, 0, 0),
            color=line_color,  # type: ignore
            parent=layer,
        )

    def _draw_slots(self, x_min: int, y_min: int, x_max: int, y_max: int, line_color: Sequence[int], space_color: Sequence[int], pad_color: Sequence[int], ):
        layer = self._drawlayer
        x1_pad, y1_pad, x2_pad, y2_pad = self._offsets
        x_space = self.cols._spacing / 2
        y_space = self.rows._spacing / 2
        cont_x_min = x_min + x1_pad
        cont_y_min = y_min + y1_pad
        cont_x_max = x_max - x2_pad
        cont_y_max = y_max - y2_pad

        for col in self.cols:
            col_state = col._state
            cell_x_min  = x_min + col_state['pos']
            cell_x_max  = cell_x_min + col_state['size']
            cell_x1_pad, cell_x2_pad = col_state['padding']

            for row in self.rows:
                row_state = row._state
                cell_y_min  = y_min + row_state['pos']
                cell_y_max  = cell_y_min + row_state['size']
                cell_y1_pad, cell_y2_pad = row_state['padding']

                if cell_x1_pad:
                    dearpygui.draw_rectangle(
                        (cell_x_min              , cell_y_max              ),
                        (cell_x_min + cell_x1_pad, cell_y_min + cell_y1_pad),
                        color=(0, 0, 0, 0),
                        fill=pad_color,  # type: ignore
                        parent=layer,
                    )
                if cell_y1_pad:
                    dearpygui.draw_rectangle(
                        (cell_x_min              , cell_y_min              ),
                        (cell_x_max - cell_x2_pad, cell_y_min + cell_y1_pad),
                        color=(0, 0, 0, 0),
                        fill=pad_color,  # type: ignore
                        parent=layer,
                    )
                if cell_x2_pad:
                    dearpygui.draw_rectangle(
                        (cell_x_max              , cell_y_min              ),
                        (cell_x_max - cell_x2_pad, cell_y_max - cell_y2_pad),
                        color=(0, 0, 0, 0),
                        fill=pad_color,  # type: ignore
                        parent=layer,
                    )
                if cell_y2_pad:
                    dearpygui.draw_rectangle(
                        (cell_x_max              , cell_y_max              ),
                        (cell_x_min + cell_x1_pad, cell_y_max - cell_y2_pad),
                        color=(0, 0, 0, 0),
                        fill=pad_color,  # type: ignore
                        parent=layer,
                    )

                dearpygui.draw_rectangle(
                    (cont_x_min, cell_y_min          ),
                    (cont_x_max, cell_y_min - y_space),
                    color=(0, 0, 0, 0),
                    fill=space_color,  # type: ignore
                    parent=layer,
                )
                dearpygui.draw_rectangle(
                    (cont_x_min, cell_y_max          ),
                    (cont_x_max, cell_y_max + y_space),
                    color=(0, 0, 0, 0),
                    fill=space_color,  # type: ignore
                    parent=layer,
                )
                # lines
                dearpygui.draw_line(
                    (cont_x_min, cell_y_min),
                    (cont_x_max, cell_y_min),
                    color=line_color,  # type: ignore
                    parent=layer,
                )
                dearpygui.draw_line(
                    (cont_x_min, cell_y_max),
                    (cont_x_max, cell_y_max),
                    color=line_color,  # type: ignore
                    parent=layer,
                )

            # outer spacing
            dearpygui.draw_rectangle(
                (cell_x_min          , cont_y_min),
                (cell_x_min - x_space, cont_y_max),
                color=(0, 0, 0, 0),
                fill=space_color,  # type: ignore
                parent=layer,
            )
            dearpygui.draw_rectangle(
                (cell_x_max          , cont_y_min),
                (cell_x_max + x_space, cont_y_max),
                color=(0, 0, 0, 0),
                fill=space_color,  # type: ignore
                parent=layer,
            )
            # lines
            dearpygui.draw_line(
                (cell_x_min, cont_y_min),
                (cell_x_min, cont_y_max),
                color=line_color,  # type: ignore
                parent=layer,
            )
            dearpygui.draw_line(
                (cell_x_max, cont_y_min),
                (cell_x_max, cont_y_max),
                color=line_color,  # type: ignore
                parent=layer,
            )

    @property
    def __code__(self):
        """[get] Return `self.__call__.__code__`.

        Enables `Grid` objects to registered as Dear PyGui callbacks.
        """
        return self.__call__.__code__

    def __call__(self, *args):
        """Draw the grid, updating the size and position of any item
        attached.

        Awaits internal lock release.


        > NOTE: When overriding, redefining, or reassigning this method,
        the new method must accept variadic positional arguments for
        `Grid` objects to be callable by Dear PyGui.
        """
        with self._lock:
            self._clean_cache()

            _area_width, _area_height, area_x_pos, area_y_pos, area_visible = self._rect_getter(
                self._target
            )

            if self._show and area_visible:
                area_width = self._width or _area_width
                self._upd_slot_states(self.cols, area_width, 0)
                area_height = self._height or _area_height
                self._upd_slot_states(self.rows, area_height, 1)

                self._upd_item_states()

                if self._overlay:
                    dearpygui.delete_item(self._drawlayer, children_only=True)
                    _item_set_config(self._drawlayer, show=True)

                    area_x_max = area_x_pos + area_width
                    area_y_max = area_y_pos + area_height
                    self._draw_outline(
                        area_x_pos,
                        area_y_pos,
                        area_x_max,
                        area_y_max,
                        (150, 255, 255),
                        (150, 255, 255, 80),
                    )
                    self._draw_slots(
                        area_x_pos,
                        area_y_pos,
                        area_x_max,
                        area_y_max,
                        (150, 255, 255, 120),
                        (150, 255, 255, 255),
                        (150, 255, 255, 80),
                    )
            else:
                _item_set_config(self._drawlayer, show=False)




if __name__ == '__main__':
    dpg = dearpygui


    dpg.create_context()
    dpg.setup_dearpygui()
    dpg.create_viewport()
    dpg.show_viewport()

    main_window = dpg.add_window()
    dpg.set_primary_window(main_window, True)
    dpg.configure_item(main_window, no_scroll_with_mouse=True, no_scrollbar=True)


    grid = Grid(2, 4, main_window, overlay=True)

    with dpg.item_handler_registry() as window_hr:
        dpg.add_item_resize_handler(callback=grid)
    dpg.bind_item_handler_registry(main_window, window_hr)




    grid.rows[0].size = 28
    with dpg.child_window(border=True, parent=main_window) as menu_bar:
        ...
    grid.push(menu_bar, (0, 0), (-1, 0))


    grid.rows[1].size = 100
    with dpg.child_window(border=True, parent=main_window) as ribbon:
        ...
    grid.push(ribbon, (0, 1), (-1, 1))


    grid.cols[0].size = 300
    with dpg.window(no_title_bar=True, no_collapse=True, no_move=True) as left_view:
        for i in range(10):
            dpg.add_button(label=f"button {i}",width=-1, height=80)

        def cb_left_view_resize(sender, app_data, user_data):
            grid.cols[0].size = dpg.get_item_width(user_data)
            grid()

        with dpg.item_handler_registry() as left_view_hr:
            dpg.add_item_resize_handler(callback=cb_left_view_resize, user_data=left_view)
        dpg.bind_item_handler_registry(left_view, left_view_hr)
    grid.push(left_view, 0, 2)


    with dpg.child_window(border=True, parent=main_window, horizontal_scrollbar=True) as right_view:
        rv_grid = Grid(10, 4, right_view, width=2000, offsets=(2, 12, 2, 12), padding=(8, 8, 8, 8), overlay=True)
        dpg.add_item_resize_handler(callback=rv_grid, parent=window_hr)
        dpg.add_item_resize_handler(callback=rv_grid, parent=left_view_hr)

        rv_grid.push(dpg.add_button(label=f"content"), (0, 0), (2, 0))
        rv_grid.push(dpg.add_button(label=f"content"), (1, 1), (3, 1))
        rv_grid.push(dpg.add_button(label=f"content"), (2, 2), (4, 2))
        rv_grid.push(dpg.add_button(label=f"content"), (3, 3), (5, 3))
        rv_grid.push(dpg.add_button(label=f"content"), (4, 0), (6, 0))
        rv_grid.push(dpg.add_button(label=f"content"), (5, 1), (7, 1))
        rv_grid.push(dpg.add_button(label=f"content"), (6, 2), (8, 2))
        rv_grid.push(dpg.add_button(label=f"content"), (7, 3), (9, 3))
    grid.push(right_view, (1, 2), (1, 2))


    grid.rows[-1].size = 24
    with dpg.child_window(border=True, parent=main_window) as footer:
        t = dpg.add_text("hello_world")
    grid.push(footer, (0, -1), (-1, -1))


    while dpg.is_dearpygui_running():
        dpg.render_dearpygui_frame()
        grid()



import types
import inspect
from inspect import Parameter
from typing import Any, Mapping, get_overloads
from pathlib import Path
from . import _typing, _interface, _mkstub
from ._mkstub import indent




try:
    DPX_PATH = Path(__file__).parent.resolve()
except AttributeError:
    DPX_PATH = Path().resolve()

typing_imports = _mkstub.Imported(_typing, False)
typing_imports.extend((
    _typing.Item,
    _typing.ItemAlias,
    _typing.ItemUUID,
    _typing.ItemCommand,
    _typing.Array,
    _typing.Literal,
    _typing.Sequence,
    _typing.Callable,
    _typing.Unpack,
    _typing.Property,
    _typing.Any,
    _typing.Color,
    _typing.Point,
    _typing.overload,
))

interface_imports = _mkstub.Imported(_interface)
interface_imports.extend((
    o for o in _mkstub.Exported.fetch(_interface).values()
    if isinstance(o, type)
    and issubclass(o, _interface.AppItemType)
))




def fetch_exports(module: types.ModuleType):
    return _mkstub.Exported.fetch(module)


def write_stub(filename: Path | str, source: str) -> None:
    #mkstub.validate_source(source)
    with open(DPX_PATH / filename, 'w') as typestub:
        typestub.write(source)




def _to_source_itp_base(tp: type[_interface.AppItemType]) -> str:
    pyi = [_mkstub.to_source_classdef(tp)]

    with indent:
        for name, anno in getattr(tp, 'configure').__annotations__.items():
            value = getattr(tp, name, Parameter.empty)
            if value is Parameter.empty or not hasattr(value, '__get__'):
                continue
            anno = _mkstub.object_annotation(anno)
            pyi.append(
                indent(f"{name}: {_typing.Property.__qualname__}[{anno}] = ...")
            )

        init_overloads = get_overloads(tp.__init__)
        if init_overloads:
            for fn in init_overloads:
                init_params = inspect.signature(fn).parameters
                init_p_src  = _mkstub.to_source_parameters(init_params)
                pyi.extend((
                    indent(f'@overload'),
                    indent(f"def __init__({_mkstub.to_source_parameters(init_params)}) -> None: ..."),
                ))

        else:
            init_params = inspect.signature(tp.__init__).parameters
            init_p_src  = _mkstub.to_source_parameters(init_params)
            icmd_params = init_params.copy()
            del icmd_params['self']

            pyi.extend((
                indent(f"def __init__({init_p_src}) -> None: ..."),
                indent(f"@staticmethod"),
                indent(f"def command({_mkstub.to_source_parameters(icmd_params)}) -> Item: ..."),
            ))

        configure_params = inspect.signature(tp.configure).parameters
        configuration_rt = _mkstub.object_annotation(
            inspect.signature(tp.configuration).return_annotation
        )
        pyi.extend((
            indent(f"def configure({_mkstub.to_source_parameters(configure_params)}) -> None: ..."),
            indent(f"def configuration(self) -> {configuration_rt}: ..."),
        ))
    return '\n'.join(pyi)


def items_pyi(fpath: str | Path):
    from . import items

    itp_imports = interface_imports.copy()
    itp_imports.append(_interface.mvAll, export=True)
    itp_imports.extend((
        o for o in fetch_exports(_interface).values()
        if not (isinstance(o, type) and issubclass(o, _interface.AppItemType))
    ), export=True)

    pyi_imp = '\n'.join((
        str(typing_imports),
        str(itp_imports),
        '\n\n\n\n'
    ))

    pyi_src1 = []
    pyi_src2 = []
    exports = fetch_exports(items)
    # TODO: use new '__aliased__' member in exports
    for name, itp in exports.items():
        if name == itp.__qualname__:
            pyi_src1.append(_to_source_itp_base(itp))
        else:
            pyi_src2.append(f"{name} = {itp.__qualname__}")
    pyi_src2.sort()

    pyi = ''.join((
        pyi_imp,
        '\n\n\n'.join(pyi_src1),
        "\n\n\n\n",
        '\n'.join(pyi_src2),
    ))
    write_stub(fpath, pyi)


def _theme_element_pyi(fpath: str | Path, module: types.ModuleType):
    from . import items

    item_imports = _mkstub.Imported(items)
    item_imports.extend((
        items.mvThemeColor,
        'ThemeColor',
        items.mvThemeStyle,
        'ThemeStyle',
    ), export=True)

    pyi_imp = '\n'.join((
        str(typing_imports),
        str(interface_imports),
        str(item_imports),
    ))

    exports = fetch_exports(module)
    aliases = exports.aliased
    pyi_src1 = "\n\n\n".join(
        _mkstub.to_source_function(itp)
        for itp in exports.values()
    )
    pyi_src2 = '\n'.join(
        f"{name} = {obj.__name__}" for name, obj in aliases.items()
    )
    pyi = "".join((pyi_imp, "\n\n\n", pyi_src1, '\n\n\n', pyi_src2))
    write_stub(fpath, pyi)


def color_pyi(fpath: str | Path):
    from . import color
    _theme_element_pyi(fpath, color)


def style_pyi(fpath: str | Path):
    from . import style
    _theme_element_pyi(fpath, style)




items_pyi("items.pyi")
color_pyi("color.pyi")
style_pyi("style.pyi")



import abc
import sys
import uuid
import inspect
import functools
from inspect import Parameter
from dearpygui import dearpygui, _dearpygui
from . import api, _typing, constants, _tools, _errors, _parsing, _mkstub
from .api import Item as ItemAPI, Registry as RegistryAPI, _create_uuid
from ._tools import classproperty
from ._typing import (
    Item,
    ItemInterface,
    ItemCommand,
    ItemConfig,
    ItemState,
    null_itemtype,
    null_command,

    Color,
    Any,
    Self,
    Array,
    Callable,
    Iterator,
    Literal,
    Iterable,
    Generic,
    Mapping,
    Property,
    Sequence,
    SupportsIndex,
    overload,

    TypedDict,
    TypeVar,
    ParamSpec,

    TYPE_CHECKING,
)
if TYPE_CHECKING:
    from .items import (
        mvTheme,
        mvFont,
        mvItemHandlerRegistry,
        mvWindowAppItem,
        mvHandlerRegistry,
        mvViewportDrawlist,
        mvFontRegistry,
        mvStage,
    )




_T = TypeVar("_T")
_P = ParamSpec("_P")

_ITP = TypeVar("_ITP", bound='AppItemType')


_exported = _mkstub.Exported(__name__)




# [ METACLASS HELPERS ]

def _is_default_command(value: Any) -> bool:
    if isinstance(value, (staticmethod, classmethod)):
        value = value.__wrapped__
    return bool(value is null_command)


def _incompatible_bases(command: ItemCommand, identity: tuple[int, str], *bases: type['AppItemType | Any']) -> bool:
    """Return True if a new item type from *bases* would
    create an incompatible union between two or more Dear
    PyGui item types.
    """
    try:
        bases = tuple(b for b in bases if issubclass(b, AppItemType))
    except NameError:
        return False
    commands = {
        b.command if not isinstance(b.command, (classmethod, staticmethod))
        else b.command.__wrapped__
        for b in bases
    }
    commands.add(command)
    commands.discard(null_command)
    identities = {b.identity for b in bases}
    identities.add(identity)
    identities.discard(null_itemtype)
    return any(len(s) > 1 for s in (commands, identities))


def _register_itemtype(cls: Any, *, register: bool | None):
    if (
        register is True
        or register is not False and not cls.__qualname__.startswith("_")
    ):
        AppItemMeta.__itemtype_registry__[cls.identity[1]] = cls


def _get_itp_subclasses(*parent_cls: type[_T]) -> tuple[type[_T], ...]:
    return tuple(
        cls for cls in AppItemMeta.__itemtype_registry__.values()
        if issubclass(cls, parent_cls)
    )


def _create_init_method(cls: Any, parameters: Mapping[str, Parameter]) -> Any:
    method = _tools.create_function(
        '__init__',
        ('self', '*args', 'tag: int | str = 0', '**kwargs'),
        (
            "super(__class__, self).__init__()",
            "try:",
            "    self.command(*args, tag=self, **kwargs)",
            "except (SystemError, TypeError) as e:",
            # BUG: The patch applied in __init__.py revealed that DPG will still
            # sometimes register foreign `tag`s even when throwing an argument-related
            # `SystemError`s (the patch ensures a `tag` is always sent so they're never
            # internally generated i.e. are always 'foreign'). In fact, the item is
            # still created in cases where no arguments are actually required. In any
            # case, it scuffs "does item exist" checks. Explicit `tag`s are *usually*
            # sent without arguments, and can be used as a decent `does_item_exist`
            # substitute.
            f"    if not _errors.{_errors.does_item_exist.__name__}(self) or (not tag and (args or kwargs)):",
            f"        err_chks = (",
            f"            _errors.{_errors.err_arg_unexpected.__name__},",
            f"            _errors.{_errors.err_parent_invalid.__name__},",
            f"        )",
            f"        for err_chk in err_chks:",
            f"            err = err_chk(self, self.command, *args, **kwargs)",
            f"            if err:",
            f"                self.destroy()",
            f"                raise err from None",
            f"        self.destroy()",
            f"        raise",
        ),
        None,
        locals={
            '__class__': cls,
            '_errors'   : _errors,
        }
    )
    method.__name__      = '__init__'
    method.__qualname__  = f'{cls.__qualname__}.__init__'

    params = [Parameter('self', Parameter.POSITIONAL_OR_KEYWORD), *parameters.values()]
    if 'kwargs' not in parameters:
        params.append(Parameter('kwargs', Parameter.VAR_KEYWORD))
    method.__signature__ = inspect.Signature(
        params,
        return_annotation=None,
    )
    command = cls.command
    if command.__defaults__:
        method.__defaults__ = tuple(command.__defaults__)
    if command.__kwdefaults__:
        method.__kwdefaults__ = dict(command.__kwdefaults__)
    method.__annotations__.clear()
    for p in method.__signature__.parameters.values():
        if p.annotation is not p.empty:
            method.__annotations__[p.name] = p.annotation
    return method


def _create_configure_method(cls: Any, parameters: Mapping[str, Parameter]) -> Any:
    """Generate and return a new `.configure` method using item
    command parameters."""
    method = _tools.create_function(
        'configure',
        ('self', '**kwargs'),
        (
            "try:",
            "    configure_item(self, **kwargs)",
            "except SystemError:",
           f"    err = _errors.{_errors.err_item_nonexistant.__name__}(self, self.command, **kwargs)",
            "    if err: raise err from None",
            "    bad_kwds = ', '.join(repr(k) for k in kwargs if k not in _ITEM_CFG_KEYS)",
            "    if bad_kwds:",
            "        raise TypeError(",
            "            f'`{type(self).__qualname__}.configure()` received '",
            "            f'unexpected keyword argument(s) {bad_kwds}.'",
            "        ) from None",
            "    for k,v in kwargs.items():",
            "        try:",
            "            configure_item(self, **{k:v})",
            "        except SystemError:",
            "            raise TypeError(",
            "                f'configuration key {k!r} of {type(self).__qualname__!r} interface '",
            "                'is read-only.'",
            "            ) from None",
        ),
        None,
        locals={
            '__class__'        : cls,
            '_errors'           : _errors,
            'configure_item'   : _dearpygui.configure_item,
            '_ITEM_CFG_KEYS'   : frozenset(parameters),
        }
    )
    method.__name__      = "configure"
    method.__qualname__  = f'{cls.__qualname__}.configure'
    method.__signature__ = inspect.Signature(
        parameters=[
            Parameter('self', Parameter.POSITIONAL_OR_KEYWORD),
            *(p.replace(default=...) for p in parameters.values()),
            Parameter('kwargs', Parameter.VAR_KEYWORD),
        ],
        return_annotation=None,
    )
    method.__annotations__.clear()
    for p in method.__signature__.parameters.values():
        if p.annotation is not p.empty:
            method.__annotations__[p.name] = p.annotation
    return method


def _create_configuration_method(cls: Any, parameters: Mapping[str, Parameter]) -> Any:
    """Generate and return a new `.configuration` method using item
    command parameters."""
    return_annotation = eval(
        f'dict[Literal[{", ".join(repr(p) for p in parameters)}], Any]',
        globals(),
        locals(),
    )
    method = _tools.create_function(
        'configuration',
        ('self',),
        (
            f"try:",
            f"    return {{",
            f"        k:v for k,v in get_item_configuration(self).items()",
            f"        if k in _ITEM_CFG_KEYS",
            f"    }}",
            f"except SystemError:",
            f"    err = _errors.{_errors.err_item_nonexistant.__name__}(self, self.command)",
            f"    if err: raise err from None",
            f"    raise",
        ),
        return_annotation,
        locals={
            '__class__'             : cls,
            '_errors'                : _errors,
            'get_item_configuration': _dearpygui.get_item_configuration,
            '_ITEM_CFG_KEYS'        : frozenset(parameters),
        },
    )
    method.__name__      = 'configuration'
    method.__qualname__  = f'{cls.__qualname__}.configuration'
    method.__signature__ = inspect.Signature(
        parameters=[
            Parameter('self', Parameter.POSITIONAL_OR_KEYWORD),
        ],
        return_annotation=return_annotation,
    )
    method.__annotations__.clear()
    for p in method.__signature__.parameters.values():
        if p.annotation is not p.empty:
            method.__annotations__[p.name] = p.annotation
    return method


# [ ITEM META & BASE ]

_AppT = TypeVar("_AppT", bound=api.Application)
_VpT  = TypeVar("_VpT", bound=api.Viewport)
_RtT  = TypeVar("_RtT", bound=api.Runtime)

class _GlobalStateObject(Generic[_T]):
    __slots__ = ("_key",)

    def __init__(self, key: str = '', /):
        self._key = key

    def __set_name__(self, cls: Any, name: str):
        self._key = self._key or f"_{name}"

    def __get__(self, instance: Any, cls: Any = None) -> _T | None:
        if instance is None:
            return self  # type: ignore
        getattr(AppItemMeta, self._key, None)

    def __set__(self, instance: Any, value: _T) -> None:
        setattr(AppItemMeta, self._key, value)



class AppItemMeta(type):
    __slots__ = ()

    __itemtype_registry__ = api.Registry.__itemtype_registry__

    identity: tuple[int, str] = null_itemtype
    command : ItemCommand     = null_command

    def __new__(
        mcls,
        name     : str,
        bases    : tuple[type[object], ...],
        namespace: dict[str, Any],
        *,
        identity: tuple[int, str] | None  = None,  # type: ignore
        command : ItemCommand | None = None,  # type: ignore
        register: bool | None        = None,  # None == 'auto', True == 'force register', False == 'do not register'
        **kwargs
    ):
        command : Any = namespace.get("command", command)
        identity: Any = namespace.get("identity", identity)

        if command and not _is_default_command(command):
            if _parsing.is_contextmanager(command):
                command = command.__wrapped__
            if not _parsing.is_item_command(command):
                if callable(command):
                    raise ValueError(f'{command!r} is not a valid item command.')
                raise TypeError(f'{command!r} is not callable.')
        if identity and identity == null_itemtype:
            try:
                int_id, str_id = identity
                if not (isinstance(int_id, int) and isinstance(str_id, str)):
                    raise TypeError
            except TypeError:
                raise TypeError(
                    f'`identity` must be a 2-tuple containing an '
                    f'integer and string value respectively (got {identity!r}).'
                ) from None

        cls = super().__new__(mcls, name, bases, namespace, **kwargs)

        # XXX: `__bool__` is purposefully undefined for all interfaces. It
        # must ALWAYS point to `int.__bool__` or eval to `bool(self.tag)`
        # because the Python C-API hooks used by DPG rely on it. If it's
        # falsy, the `tag` value of the interface is treated as 0.
        if issubclass(cls, int) and getattr(cls, '__bool__') is not int.__bool__:
            raise TypeError(
                f'derived item interface types cannot override `__bool__`.'
            )

        if all((command, identity)):  # try to process as a new base item type
            cls.identity = identity
            cls.command  = (
                staticmethod(command)
                if not isinstance(command, staticmethod)
                else command
            )
            if _is_default_command(command) or identity == null_itemtype:
                # early exit for primitive bases
                return cls

            if _incompatible_bases(command, identity, *bases):
                raise TypeError(
                    f"cannot create class {name!r} -- incompatible union of bases."
                )

            tp_def = _parsing.item_definitions()[
                cls.identity[1].removeprefix("mvAppItemType::")
            ]

            # these methods are built only once per base type
            if getattr(cls, '__init__', object.__init__) is object.__init__:
                setattr(cls, '__init__', _create_init_method(cls, tp_def.init_params))
            if getattr(cls, "configure", api.Item.configure) is api.Item.configure:
                setattr(cls, 'configure', _create_configure_method(cls, tp_def.wconfig_params))
            if getattr(cls, "configuration", api.Item.configuration) is api.Item.configuration:
                setattr(cls, 'configuration', _create_configuration_method(cls, tp_def.rconfig_params))

            # Expose "configuration" options as a rw properties, even if
            # they aren't writable. The `.configure` method will handle
            # those _errors.
            config_params = list(tp_def.rconfig_params)
            config_params.extend(
                opt for opt in tp_def.wconfig_params if opt not in config_params
            )
            for opt in config_params:
                if hasattr(cls, opt):
                    continue  # `AppItemType` member, probably
                config_desc = ItemConfig()
                setattr(cls, opt, config_desc)
                config_desc.__set_name__(cls, opt)  # type: ignore

            _register_itemtype(cls, register=register)

        elif register is True:
            _register_itemtype(cls, register=register)

        return cls

    def __repr__(self) -> str:
        return f"<class {self.__qualname__!r}>"

    def __hash__(self):
        return hash((self.__name__, self.__module__, *self.identity))

    def __eq__(self, other: Any):
        # Being very explicit w/*other* typing here to prevent
        # unexpected results in niche cases.
        if isinstance(other, str) and other.startswith("mvAppItemType::"):
            return self.identity[1] == other
        if type(other) is int:
            return self.identity[0] == other
        return other is self

    def __str__(self) -> str:
        return self.identity[1]

    def __int__(self) -> int:
        return self.identity[0]

    application: Property[_AppT | None] = _GlobalStateObject()  # type: ignore
    viewport   : Property[_VpT  | None] = _GlobalStateObject()  # type: ignore
    runtime    : Property[_RtT  | None] = _GlobalStateObject()  # type: ignore

    def start(self, *args, **kwargs) -> Any:
        """Start the user interface event loop, blocking the thread until
        the viewport is closed.

        This method forwards the call to `self.runtime.start`. If a
        `Runtime` object has not yet been set, an internal default
        `.start` method is called instead. The default implementation
        performs all required setup, if not done so already.
        """
        rt = self.runtime
        if rt is None:
            rt = api.Runtime
        return rt.start(*args, **kwargs)

    def stop(self, *args, **kwargs) -> Any:
        """Close the viewport and kill the runtime.

        This method forwards the call to `self.runtime.stop`. If a
        `Runtime` object has not yet been set, an internal default
        `.stop` method is called instead.
        """
        rt = self.runtime
        if rt is None:
            rt = api.Runtime
        return rt.stop(*args, **kwargs)


class ABCAppItemMeta(AppItemMeta, abc.ABCMeta):
    """Metaclass for creating abstract app item interface types."""









# [ `AppItemType` HELPERS ]

def _onetime_setup(fn: _T) -> _T:
    """`AppItemType.__new__ wrapper`. Performs application-level setup
    once.
    """
    @functools.wraps(fn)  # type: ignore
    def application_setup(*args, **kwargs):
        if not api.Application.state()['ok']:
            api.Application.create_context()
            api.Application.prepare()
        obj = fn(*args, **kwargs)  # type: ignore
        AppItemType.__new__ = fn  # type: ignore
        return obj
    return application_setup  # type: ignore


def _get_base_itp(item: Item, information: _typing.ItemInfoDict | None = None):
    return AppItemMeta.__itemtype_registry__[
        (information or ItemAPI.information(item))['type']
    ]


# `AppItemType.__reduce__`

_Reduced = tuple[
    Callable[..., _ITP],                                     # `_new_from_reduced`
    tuple[type[_ITP], str, tuple[Any, ...], dict[str, Any]], # (class, alias, args, kwargs)
    '_SaveState',                                            # ...
    Iterator[Any] | None,                                    # None
    Iterator[tuple[str, Any]] | None,                        # None
    Callable[[_ITP, '_SaveState'], _ITP],                    # `_load_reduced`
]

class _SaveState(TypedDict):
    alias         : str
    type          : str
    parent        : int | str | None
    theme         : Any
    font          : Any
    handlers      : Any
    children      : tuple[list[Any], list[Any], list[Any], list[Any]]
    configuration : dict[str, Any]
    value         : Any
    itf_state     : dict[str, Any] | None
    itf_state_keys: tuple[str, ...]

def _to_save_state(item: 'AppItemType', info: _typing.ItemInfoDict | None = None, **kwargs) -> _SaveState:
    info   = info or item.information()
    config = item.configuration()
    pos    = item.state()['pos']
    if pos and any(v > 0 for v in pos):
        config['pos'] = pos
    config.update(kwargs)

    return {
        "alias"         : item.get_alias(),
        "type"          : info["type"],
        "parent"        : info["parent"],
        "theme"         : info["theme"],
        "font"          : info["font"],
        "handlers"      : info["handlers"],
        "children"      : tuple(info['children'].values()),
        "configuration" : config,                             # type: ignore
        "value"         : item.get_value(),
        # these are set by things that use them
        "itf_state"     : None,
        "itf_state_keys": (),
    }

def _new_from_reduced(cls: type['AppItemType'], alias: str, *args, **kwargs):
    if RegistryAPI.item_exists(alias):
        kwargs['tag'] = alias
    return cls.__new__(cls, *args, **kwargs)

def _load_reduced(item: 'AppItemType', save_state: _SaveState, source_map: dict['AppItemType', str] | None = None, root_caller: bool = True):
    """Loader for nested reduced/pickled item interfaces."""
    if root_caller:
        assert source_map is None
        source_map = {}

    # Re-create the item state. The check sort of acts
    # as a memo map; reducing the chances of creating
    # interfaces of items already re-created.
    alias = save_state['alias']
    if not RegistryAPI.item_exists(alias):
        config = save_state['configuration']
        source = config.pop('source', 0)
        parent = save_state['parent']
        if parent:
            item.command(tag=item, parent=parent, **config)
        else:
            item.command(tag=item, **config)
        item.set_alias(alias)

        for k in ('theme', 'font', 'handlers'):
            reduced = save_state[k]
            if reduced:
                newfn, newargs, ss, *_ = reduced
                newcls, _alias, newargs, newkwds = newargs
                itf = newfn(newcls, _alias, *newargs, **newkwds)
                _load_reduced(itf, ss, source_map, False)
                getattr(item, f'set_{k}')(itf)

        for slot in save_state['children']:
            for reduced in slot:
                newfn, newargs, ss, *_ = reduced
                newcls, _alias, newargs, newkwds = newargs
                itf = newfn(newcls, _alias, *newargs, **newkwds)
                ss['parent'] = alias
                _load_reduced(itf, ss, source_map, False)

        item.set_value(save_state['value'])

        config = save_state['configuration']

        if source:
            # don't set it now -- the root caller will try later
            if not RegistryAPI.item_exists(source):
                source_map[item] = source  # type: ignore
            else:
                ItemAPI.configure(item, source=source)

    itf_state = save_state['itf_state']
    if itf_state:
        for k in save_state['itf_state_keys']:
            newfn, newargs, ss, *_ = itf_state[k]
            newcls, _alias, newargs, newkwds = newargs
            itf = newfn(newcls, _alias, *newargs, **newkwds)
            _load_reduced(itf, ss, source_map, False)
            itf_state[k] = itf
        setstate = getattr(item, '__setstate__', None)
        if setstate:
            setstate(itf_state)
        else:
            for k, v in itf_state.items():
                setattr(item, k, v)

    # If sources do not exist at this point, they were not
    # referenced in any identifiable way in the reduced tree.
    if root_caller:
        for itf, source in source_map.items():    # type: ignore
            try:
                itf.configure(source=source)
            except SystemError:
                if not RegistryAPI.item_exists(source):
                    raise RuntimeError(
                        f'{type(item).__name__!r} state loading error -- '
                        f'`source` item {source!r} does not exist.'
                    )
                raise
    return item



# Interfaces need to be instances of the built-in `int` or `str` type
# to allow them to be used for `tag`, `item`, etc. arguments. This is
# because Dear PyGui uses `PyLong_Check` and `PyUnicode_Check` which
# takes a shortcut; checking the class' `tp_flags` instead of the
# normal `isinstance/issubclass` procedure.
@ItemInterface.register
class AppItemType(api.Item, int, register=False, metaclass=AppItemMeta):
    __slots__ = ()

    # [ CONSTRUCTORS ]


    @_onetime_setup
    def __new__(cls, *args, tag: Item = 0, **kwargs) -> Self:
        """Args:
            * tag: Item identifier, in-use or otherwise. If 0 or None, an
            integer identifier is automatically generated.

        NOTE: If an item would be created as a result of creating an
        interface, *tag* should not be a string alias.
        """
        if not tag:
            tag = _create_uuid()
        elif isinstance(tag, str):
            tag = RegistryAPI.get_item_uuid(tag)
            if not tag:
                # DPG's item registry API was never finished, so it's impossible
                # to register an alias and tell DPG to actually use the uuid
                # bound to the alias for the item. Basically; manually registering
                # aliases does absolutely nothing. A work-around is complicated,
                # so just use the alternate constructor.
                raise ValueError(
                    "using use the `.aliased` method "
                )
        return super().__new__(cls, tag)

    @classmethod
    def aliased(cls, *args, tag: Item, **kwargs) -> Self:
        """Alternate interface constructor that supports string aliases
        when creating both an interface and item.

        Args:
            * tag: Item identifier, in-use or otherwise. If 0 or None, an
            integer identifier is automatically generated.
        """
        if not isinstance(tag, str):
            return cls(*args, tag=tag, **kwargs)
        self = cls(*args, **kwargs)
        self.set_alias(tag)
        return self

    @classmethod
    def new(cls, tag: Item = 0) -> Self:
        """Return a new interface. The instance is not initialized.

        Args:
            * tag: Item identifier, in-use or otherwise. If 0 or None, an
            integer identifier is automatically generated.


        If the instance's tag value is unused, expect _errors when using
        most of the instance-bound API. The instance can be initialized
        at any time by calling the `.__init__` or `.init` methods.

        The new instance returned by this method is created by directly
        calling the class' `__new__` method (using *args* and *kwargs*)
        without ever calling `__init__` and NOT through typical
        invocation.

        It is not necessary to hook through this method to create an
        interface for an existing item, since the default `__init__`
        implementation will handle the resulting exception. This method
        is more performant over the former when creating interfaces for
        several existing items.
        """
        return cls.__new__(cls, tag=tag)

    def init(self, *args, **kwargs):
        """Initialize or re-initialize the interface. If the instance's
        tag value is unused, a proper item will be created to validate
        the interface's existance.

        Args:
            * args: Positional arguments used to create the item.

            * kwargs: Keyword arguments used to create the item.


        Required arguments can be omitted if the interface's tag value
        is already in-use.
        """
        return self.__init__(*args, **kwargs)


    # [ BEHAVIORS ]

    def __repr__(self):
        return (
            f"{type(self).__qualname__}("
            f"tag={int(self)!r}, "
            f"alias={ItemAPI.get_alias(self)!r}, "
            f"label={ItemAPI.configuration(self)['label']!r}, "
            f"parent={ItemAPI.information(self)['parent']!r}"
            f")"
        )

    @overload
    def __getitem__(self, index: Literal[0, 1, 2, 3]) -> list[Item]: ...
    @overload
    def __getitem__(self, index: slice) -> list[list[Item]]: ...
    def __getitem__(self, index: Literal[0, 1, 2, 3] | slice) -> list[Item] | list[list[Item]]:
        slots = self.information()["children"]
        try:
            return slots[index]  # type: ignore
        except KeyError:
            raise IndexError(f"expected slot index, got {index!r}.") from None
        except TypeError:  # unhashable
            if isinstance(index, slice):
                return [*slots.values()][index.start: index.stop: index.step]  # type: ignore
            raise

    def __getnewargs_ex__(self):
        return (), {}

    def __reduce_ex__(self, protocol: int = -1) -> _Reduced:
        save_state = _to_save_state(self)

        alias = save_state['alias']
        if not alias:
            alias = save_state['alias'] = RegistryAPI.create_alias(self.tag)
            self.set_alias(alias)

        for k in ('theme', 'font', 'handlers'):
            item = save_state[k]
            if item:
                save_state[k] = _get_base_itp(item).new(item).__reduce_ex__(protocol)

        for slot in save_state['children']:
            for i, child in enumerate(slot):
                slot[i] = _get_base_itp(child).new(child).__reduce_ex__(protocol)

        # When unpickled, the loader will inject the parent's
        # alias into a child's save state before loading it.
        # Clearing the existing parent here simply allows for
        # the top-level item take advantage of the container
        # stack when it is not a root item i.e.
        # `with Window(): pickle.load(...)`.
        save_state['parent'] = 0

        # `source` is extremely tricky to handle since it may
        # exist in a separate item branch. Ensure it has an alias
        # and hope for the best. Ideally, the user will pickle it
        # as well, and load beforehand. No worries if `source` is
        # a child of this item, since the loader can cover that.
        source = save_state['configuration'].get('source', 0)
        if source:
            _source = ItemAPI.get_alias(source)
            if not _source:
                _source = f'{source}-{uuid.uuid4()}'
                ItemAPI.set_alias(source, _source)
            save_state['configuration']['source'] = _source

        itf_state = save_state['itf_state'] = self.__getstate__()
        if itf_state:
            itf_state_keys = []
            for k, v in itf_state.items():
                # it falls upon the user to serialize anything else
                if isinstance(v, AppItemType):
                    itf_state[k] = v.__reduce_ex__(protocol)
                    itf_state_keys.append(k)
            save_state['itf_state_keys'] = tuple(itf_state_keys)

        cls = self.__class__
        return (
            _new_from_reduced,
            (cls, alias, *self.__getnewargs_ex__()),
            save_state,
            None,
            None,
            _load_reduced,
        )

    __getstate__: Callable[..., dict[str, Any]]

    @overload
    def __copy__(self, **kwargs) -> Self: ...
    @overload
    def __copy__(self) -> Self: ...
    def __copy__(self, __save_state__: _SaveState | None = None, **kwargs) -> Self:
        copy_state = __save_state__ or _to_save_state(self, **kwargs)

        nargs, nkwds = self.__getnewargs_ex__()
        copy_item = self.__class__.__new__(self.__class__, *nargs, **nkwds)
        copy_item.command(tag=copy_item, **copy_state['configuration'])

        for k in ('font', 'theme', 'handlers'):
            item = copy_state[k]
            if item:
                getattr(copy_item, f"set_{k}")(item)

        if self.is_container:
            c_parent = copy_item
        else:
            c_parent = self
        for slot in copy_state['children']:
            for child in slot:
                child_info = ItemAPI.information(child)
                child_itf  = _get_base_itp(child).new(child)
                child_ss   = _to_save_state(child_itf, child_info)
                child_ss['configuration']['parent'] = c_parent
                child_itf.__copy__(parent=c_parent)

        itf_state = copy_state['itf_state']
        if itf_state:
            setstate = getattr(copy_item, '__setstate__', None)
            if setstate:
                setstate(copy_item, itf_state)
            else:
                for k,v in copy_state.items():
                    setattr(copy_item, k, v)

        return copy_item

    def __deepcopy__(self, memo: Any = None):
        raise NotImplementedError


    # [ GLOBAL STATE API ]

    # These are dependant on hooks being defined in two places; on
    # the metaclass, and below. They are read/write at the (non-meta)
    # class level, but read-only at the instance level. It falls upon
    # the user to set these at runtime, if they want. Otherwise they'll
    # simply return None.

    @property
    def application(self):
        """[get] Return the global `Application` object."""
        return AppItemMeta.application

    @property
    def viewport(self):
        """[get] Return the global `Viewport` object."""
        return AppItemMeta.viewport

    @property
    def runtime(self):
        """[get] Return the global `Runtime` object."""
        return AppItemMeta.runtime


    # [ CLASS-BOUND API ]

    identity: tuple[int, str] = null_itemtype
    command : ItemCommand     = null_command    # metaclass -> command = staticmethod(command)

    @classproperty
    def is_container(cls) -> bool:
        """[get] Return True if items of this type can contain other
        non-root items."""
        # A more accurate check would be to call `is_item_container`,
        # but that would require an existing item.
        return issubclass(cls, ContainerType)  # type: ignore

    @classproperty
    def is_root_item(cls) -> bool:
        """[get] Return True if items of this type cannot be parented.
        """
        return issubclass(cls, RootType)  # type: ignore

    @classproperty
    def is_plot_item(cls) -> bool:
        """[get] Return True if items of this type are included in Dear
        PyGui's plotting API."""
        return issubclass(cls, PlottingType)  # type: ignore

    @classproperty
    def is_node_item(cls) -> bool:
        """[get] Return True if items of this type are included in Dear
        PyGui's node API."""
        return issubclass(cls, NodeType)  # type: ignore

    @classproperty
    def is_theme_item(cls) -> bool:
        """[get] Return True if items of this type are included in Dear
        PyGui's theming API.
        """
        return issubclass(cls, ThemeType)  # type: ignore

    @classproperty
    def is_table_item(cls) -> bool:
        """[get] Return True if items of this type are included in Dear
        PyGui's table API."""
        return issubclass(cls, TableType)  # type: ignore

    @classproperty
    def is_draw_item(cls) -> bool:
        """[get] Return True if items of this type are included in Dear
        PyGui' drawing API."""
        return issubclass(cls, DrawingType)  # type: ignore

    @classproperty
    def target_slot(cls) -> Literal[0, 1, 2, 3]:
        """[get] Return the index of the child slot that would store
        items of this type.
        """
        # This is included as the 'target' key in the return value of
        # `get_item_information()`. But that requires an item to actually
        # exist.
        tp_name = cls.identity[1]
        if tp_name in (
            'mvAppItemType::mvFileExtension',
            'mvAppItemType::mvFontRangeHint',
            'mvAppItemType::mvNodeLink',
            'mvAppItemType::mvAnnotation',
            'mvAppItemType::mvDragLine',
            'mvAppItemType::mvDragPoint',
            'mvAppItemType::mvLegend',
            'mvAppItemType::mvTableColumn',
        ):
            slot = 0
        elif tp_name == 'mvAppItemType::mvDragPayload':
            slot = 3
        elif tp_name.startswith("mvAppItemType::mvDraw"):
            slot = 2
        else:
            slot = 1  # for root items, too
        return slot


    # [ INSTANCE-BOUND API ]

    @property
    def tag(self) -> int:
        """[get] Return the item's unique integer identifier."""
        return self.real

    @property
    def alias(self) -> str:
        """[get] Return the item's unique string identifier."""
        return self.get_alias()
    @alias.setter
    def alias(self, value: str | None) -> None:
        """[set] Set a unique string identifier for the item."""
        if value is None:
            alias = self.alias
            if alias is None:
                return
            return RegistryAPI.remove_alias(alias)
        return self.set_alias(value)

    # configuration members
    label             : Property[str | None] = ItemConfig()
    user_data         : Property[Any]        = ItemConfig()
    use_internal_label: Property[bool]       = ItemConfig()

    # information members
    @property
    def parent(self) -> 'mvAll':
        """[get] Return an interface for the item's parent, or None if
        the item is not parented.
        """
        parent = self.information()['parent']
        return mvAll(parent) if parent else parent  # type: ignore

    @property
    def theme(self) -> 'mvTheme | Item | None':
        """[get] Return the theme set onto the item."""
        theme = self.get_theme()
        try:
            return AppItemMeta.__itemtype_registry__['mvAppItemType::mvTheme'].new(
                theme
            ) if theme else theme
        except KeyError:
            return theme
    @theme.setter
    def theme(self, value: Item | None) -> None:
        """[set] Set the reflected theme item."""
        self.set_theme(value)

    @property
    def font(self) -> 'mvFont | Item | None':
        """[get] Return the font item in use."""
        font = self.get_font()
        try:
            return AppItemMeta.__itemtype_registry__['mvAppItemType::mvFont'].new(
                font
            ) if font else font
        except KeyError:
            return font
    @font.setter
    def font(self, value: Item) -> None:
        """[set] Set the reflected font item."""
        self.set_font(value)

    @property
    def handlers(self) -> 'mvItemHandlerRegistry | Item | None':
        """[get] Return the handler registry used by the item."""
        handlers = self.get_handlers()
        try:
            return AppItemMeta.__itemtype_registry__['mvAppItemType::mvItemHandlerRegistry'].new(
                handlers
            ) if handlers else handlers
        except KeyError:
            return handlers
    @handlers.setter
    def handlers(self, value: Item | None) -> None:
        """[set] Set an item handler registry for the item to use."""
        self.set_handlers(value)

    # state members
    is_ok                    : Property[bool | None]            = ItemState("ok")
    is_hovered               : Property[bool | None]            = ItemState("hovered")
    is_active                : Property[bool | None]            = ItemState("active")
    is_focused               : Property[bool | None]            = ItemState("focused")
    is_clicked               : Property[bool | None]            = ItemState("clicked")
    is_left_clicked          : Property[bool | None]            = ItemState("left_clicked")
    is_right_clicked         : Property[bool | None]            = ItemState("right_clicked")
    is_middle_clicked        : Property[bool | None]            = ItemState("middle_clicked")
    is_visible               : Property[bool | None]            = ItemState("visible")
    is_edited                : Property[bool | None]            = ItemState("edited")
    is_activated             : Property[bool | None]            = ItemState("activated")
    is_deactivated           : Property[bool | None]            = ItemState("deactivated")
    is_deactivated_after_edit: Property[bool | None]            = ItemState("deactivated_after_edit")
    is_resized               : Property[bool | None]            = ItemState("resized")
    rect_min                 : Property[Array[int, int] | None] = ItemState()
    rect_max                 : Property[Array[int, int] | None] = ItemState()
    rect_size                : Property[Array[int, int] | None] = ItemState()
    content_region_avail     : Property[Array[int, int] | None] = ItemState()

    # misc
    @property
    def value(self) -> Any:
        """[get] Return the item's stored value."""
        return self.get_value()
    @value.setter
    def value(self, value: Any) -> None:
        """[set] Set the item's value. Not all items can hold value."""
        self.set_value(value)

    @property
    def root_parent(self):
        """[get] Return an interface for the top-level parent of this item
        branch."""
        root_parent = super().root_parent()
        return self if self == root_parent else interface(root_parent)



class ABCAppItemType(AppItemType, metaclass=ABCAppItemMeta):
    """Abstract interface base class."""
    __slots__ = ()



@_exported
class mvAll(
    api.Window,  # needs to be before Table
    api.Table,
    api.Drawing,
    api.Plot,
    api.NodeEditor,
    api.Container,
    api.ThemeElement,
    api.BasicItem,
    AppItemType,
):
    """Generic interface for Dear PyGui items.

    Unlike other interface types, `mvAll` cannot create new
    items. Instead, its' constructor is optimized for existing
    uuids and aliases.

    Instances of `mvAll` have an expansive API to accomodate a
    wide range of items. However, they lack type identity and
    do not implement behavior methods unique to other
    interfaces. They can still be used as context managers;
    a `TypeError` is raised if the interfaced item is not a
    container.

    The `mvAll` type has an identity integer value of 0.
    """
    """Generic item interface with a constructor optimized for existing
    items. It features an expansive API, but lacks type identity and
    behavior unique to specific interface types.

    Unlike the typical item type constructor, `mvAll` will not generate
    any new identifiers. It accepts a sole (required) *tag* argument,
    with a strong preference of `int` identifiers.

    If *tag* is a `str` alias, it will try to fetch the related uuid.

    `mvAll` is the fallback interface for the `interface` function.

    The `mvAll` type has an integer value of 0.
    """
    __slots__ = ()

    def __new__(cls, tag: Item) -> Self:
        # I've profiled almost every realistic implementation
        # under the sun trying to get this to be as fast as
        # possible. One exists, but the gain is EXTREMELY
        # marginal. Even then, it would only benefit integer
        # `tag` values.
        if isinstance(tag, int):
            return int.__new__(cls, tag)
        return int.__new__(cls, RegistryAPI.get_item_uuid(tag))

    def __init__(self, tag: Item): ...

    def __enter__(self) -> Self:
        try:
            self.push_stack()
        except:
            if RegistryAPI.item_exists(self):
                raise TypeError(
                    f"{type(self).__qualname__!r} is not a container item."
                )
            raise
        return self

    def __exit__(self, *args) -> None:
        self.pop_stack()

    def __reduce_ex__(self, protocol: int = -1):
        return _get_base_itp(self).new(self).__reduce_ex__(protocol)

    # Identity-related class properties would always return False,
    # so they need to be re-implemented. Fortunately, class-level
    # insight isn't necessary.

    @property
    def is_container(self) -> bool:
        """[get] Return True if this item can contain other
        non-root items."""
        return self.information()['container']

    @property
    def is_root_item(self) -> bool:
        """[get] Return True if this item cannot be parented.
        """
        return self.information()['parent'] is not None

    @property
    def is_plot_item(self) -> bool:
        """[get] Return True if this item is included in Dear
        PyGui's plotting API."""
        tp_name = self.information()['type'].split("::", maxsplit=1)[-1]
        return tp_name.startswith('mvPlot') or tp_name == 'mvAnnotation'

    @property
    def is_node_item(self) -> bool:
        """[get] Return True if this item is included in Dear
        PyGui's node API."""
        tp_name = self.information()['type'].split("::", maxsplit=1)[-1]
        return tp_name.startswith('mvNode')

    @property
    def is_theme_item(self) -> bool:
        """[get] Return True if this item is included in Dear
        PyGui's theming API.
        """
        tp_name = self.information()['type'].split("::", maxsplit=1)[-1]
        return tp_name.startswith('mvTheme')

    @property
    def is_table_item(self) -> bool:
        """[get] Return True if this item is included in Dear
        PyGui's table API."""
        tp_name = self.information()['type'].split("::", maxsplit=1)[-1]
        return tp_name.startswith('mvTable')

    @property
    def is_draw_item(self) -> bool:
        """[get] Return True if this item is included in Dear
        PyGui' drawing API."""
        tp_name = self.information()['type'].split("::", maxsplit=1)[-1]
        return tp_name.startswith('mvDraw') or "Drawlist" in tp_name




# [ COMPONENT TYPES ]

# LOW-LEVEL CLASSIFICATION TYPES

@_exported
class BasicType(api.BasicItem, AppItemType):
    """Interface for Dear PyGui's basic non-container items."""
    __slots__ = ()

    root_parent: Property['RootType']


@_exported
class ContainerType(api.Container, AppItemType):
    """Interface for Dear PyGui items that can parent other
    items.

    Items that can be parented vary based on the item's type.
    """
    __slots__ = ()

    def __enter__(self) -> Self:
        self.push_stack()
        return self

    def __exit__(self, *args) -> None:
        self.pop_stack()

    @property
    def is_top_stack(self):
        return super().is_top_stack()


@_exported
class RootType(ContainerType):
    """Interface for top-level Dear PyGui container items.
    Root items cannot be parented.
    """
    __slots__ = ()

    parent     : Property[None]
    root_parent: Property[Self]


@_exported
class RegistryType(RootType):
    """Interface for hidden, non-interactable, top-level
    Dear PyGui items."""
    __slots__ = ()



# HIGH-LEVEL CLASSIFICATION TYPES

@_exported
class HandlerType(api.BasicItem, AppItemType):
    """Interface for basic Dear PyGui items that fire a
    callback given a certain event. They are always parented
    by a `RegistryType` item.
    """
    __slots__ = ()

    root_parent: Property['mvHandlerRegistry | mvItemHandlerRegistry | mvStage']

    MouseInput = constants.MouseInput
    KeyInput   = constants.KeyInput


@_exported
class NodeType(AppItemType):
    __slots__ = ()

@_exported
class NodeEditorType(api.NodeEditor, NodeType):
    __slots__ = ()


@_exported
class ThemeType(AppItemType):
    __slots__ = ()

    root_parent: Property['mvTheme']

@_exported
class ThemeElementType(api.ThemeElement, ThemeType):
    __slots__ = ()



@_exported
class FontType(AppItemType):
    __slots__ = ()

    root_parent: Property['mvFontRegistry']

    def text_size(self, text: str, *, wrap_width: int = -1):
        return ItemAPI.text_size(text, wrap_width=wrap_width, font=self)


@_exported
class DrawingType(api.Drawing, AppItemType):
    __slots__ = ()

    root_parent: Property['mvWindowAppItem | mvViewportDrawlist | mvStage']

@_exported
class DrawNodeType(api.DrawNode, DrawingType):
    __slots__ = ()


@_exported
class PlottingType(AppItemType):
    __slots__ = ()

@_exported
class PlotType(api.Plot, PlottingType):
    __slots__ = ()

@_exported
class PlotAxisType(api.PlotAxis, PlottingType):
    __slots__ = ()


@_exported
class TableType(AppItemType):
    __slots__ = ()

@_exported
class TableItemType(api.Table, TableType):
    __slots__ = ()

    @property
    def x_scroll_max(self):
        return super().x_scroll_max()

    @property
    def x_scroll_pos(self):
        return self.get_x_scroll_pos()
    @x_scroll_pos.setter
    def x_scroll_pos(self, value: float):
        self.set_x_scroll_pos(value)

    @property
    def y_scroll_max(self):
        return super().y_scroll_max()

    @property
    def y_scroll_pos(self):
        return self.get_y_scroll_pos()
    @y_scroll_pos.setter
    def y_scroll_pos(self, value: float):
        self.set_y_scroll_pos(value)


@_exported
class WindowType(api.Window, AppItemType):
    __slots__ = ()

    @property
    def x_scroll_max(self):
        return super().x_scroll_max()

    @property
    def x_scroll_pos(self):
        return self.get_x_scroll_pos()
    @x_scroll_pos.setter
    def x_scroll_pos(self, value: float):
        self.set_x_scroll_pos(value)

    @property
    def y_scroll_max(self):
        return super().y_scroll_max()

    @property
    def y_scroll_pos(self):
        return self.get_y_scroll_pos()
    @y_scroll_pos.setter
    def y_scroll_pos(self, value: float):
        self.set_y_scroll_pos(value)

    @property
    def is_active_window(self) -> bool:
        """[get] Return True if this window or any of its' children are focused."""
        return self.tag == api.Viewport.active_window()



# FUNCTIONAL CLASSIFICATION TYPES
# (these do not inherit from a base API)

class _ItemStateSized(_typing.ItemStateDict):
    pos                 : Array[int, int]  # type: ignore
    rect_size           : Array[int, int]  # type: ignore

@_exported
class SupportsSized(AppItemType):
    """Interface for rendered items with a bounding box. They can be resized
    and explicitly positioned."""
    __slots__ = ()

    width : Property[int] = ItemConfig()
    height: Property[int] = ItemConfig()

    def state(self) -> _ItemStateSized: ...
    del state

    # Some items like `mvWindowAppItem` only support `rect_size`. The min/max
    # for these can be easily calculated.
    @property
    def rect_min(self) -> Array[int, int]:
        state = self.state()
        return state['rect_min'] or state['pos']

    @property
    def rect_max(self) -> Array[int, int]:
        state = self.state()
        if rect_max:=state['rect_max']:
            return rect_max

        x_pos, y_pos = state['pos']
        config = self.configuration()
        return x_pos + config["width"], y_pos + config["height"]

    @property
    def pos(self) -> Array[int, int]:
        return self.state()['pos']
    @pos.setter
    def pos(self, value: Array[int, int]):
        self.configure(pos=value)


@_exported
class SupportsValueArray(AppItemType, Generic[_T]):
    """Interface for items that can store an array of arbitrary size."""
    __slots__ = ()

    def get_value(self) -> list[_T]:
        return ItemAPI.get_value(self) or []

    def __add__(self, other: Iterable[_T]) -> Self:
        value = self.get_value()
        value.extend(other)
        self.set_value(value)
        return self

    __iadd__ = __add__

    def __mul__(self, other: int) -> Self:
        value = self.get_value()
        self.set_value(value * other)
        return self

    __imul__ = __mul__

    @overload
    def __getitem__(self, index: SupportsIndex, /) -> _T: ...
    @overload
    def __getitem__(self, slice: slice, /) -> list[_T]: ...
    def __getitem__(self, index: Any, /):
        return self.get_value()[index]

    @overload
    def __setitem__(self, index: SupportsIndex, value: _T, /) -> None: ...
    @overload
    def __setitem__(self, index: slice, value: Sequence[_T], /) -> None: ...
    def __setitem__(self, index: Any, /, value: Any) -> None:
        item_value = self.get_value()
        item_value[index] = value
        self.set_value(item_value)

    def __delitem__(self, index: int):
        value = self.get_value()
        del value[index]
        self.set_value(value)

    def __contains__(self, value: _T):
        return value in self.get_value()

    def __reversed__(self):
        yield from reversed(self.get_value())

    def __str__(self):
        return str(self.get_value())

    def __len__(self):
        return len(self.get_value())

    def __iter__(self):
        return iter(self.get_value())

    def __lt__(self, other: Any):
        return self.get_value() < other

    def __le__(self, other: Any):
        return self.get_value() <= other

    def __eq__(self, other: Any):
        return self.get_value() == other

    def __ge__(self, other: Any):
        return self.get_value() >= other

    def __gt__(self, other: Any):
        return self.get_value() > other

    def count(self, value: Any) -> int:
        return self.get_value().count(value)

    def index(self, value: _T, start: int = 0, stop: int = sys.maxsize) -> int:
        return self.get_value().index(value, start, stop)

    def append(self, value: _T) -> None:
        """Add *value* to the end of this item's values."""
        item_value = self.get_value()
        item_value.append(value)
        self.set_value(item_value)

    def extend(self, iterable: Sequence[_T]) -> None:
        """Append values in *iterable* to this item's values."""
        item_value = self.get_value()
        item_value.extend(iterable)
        self.set_value(item_value)

    def pop(self, index: int = -1) -> _T:
        """Remove and return the last value at *index*."""
        item_value = self.get_value()
        value = item_value.pop(index)
        self.set_value(item_value)
        return value

    def insert(self, index: int, value: _T):
        """Insert *value* before *index*."""
        item_value = self.get_value()
        item_value.insert(index, value)
        self.set_value(item_value)

    def remove(self, value: _T):
        """Remove the first occurance of *value*."""
        item_value = self.get_value()
        item_value.remove(value)
        self.set_value(item_value)

    @overload
    def sort(self, /, *, key: Callable[[_T], Any] = ..., reverse: bool = ...) -> None: ...
    @overload
    def sort(self, /) -> None: ...
    def sort(self, /, **kwargs) -> None:
        """Sort values in-place."""
        item_value = self.get_value()
        item_value.sort(**kwargs)
        self.set_value(item_value)

    def reverse(self) -> None:
        """Reverse item values in-place."""
        self.set_value(self.get_value()[::-1])


@_exported
class SupportsCallback(AppItemType, Generic[_P, _T]):
    """Interface for items that support an optional callback
    event.

    The callback is exposed via the 'callback' item
    configuration option and instance attribute. Additonally,
    calls made to the interface are forwarded to the assigned
    callback.

    Instances of `SupportsCallback` implement Dear PyPixl's
    `ItemCallback` protocol, and can be used as arguments for
    Dear PyGui `callback`-related arguments.
    """
    __slots__ = ()

    def __init_subclass__(cls):
        super().__init_subclass__()
        if '__call__' in cls.__dict__ and not '__code__' in cls.__dict__:
            setattr(cls, '__code__', getattr(cls, '__call__').__code__)

    @property
    def callback(self) -> Callable[_P, _T] | None:
        """[get] Return the item's assigned callback."""
        return self.configuration()["callback"]  # type: ignore
    @callback.setter
    def callback(self, value: Callable[_P, _T] | None) -> Callable | None:
        """[set] Set the item's callback."""
        self.configure(callback=value)
        return value

    def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> None:
        callback = self.callback
        if callback:
            callback(*args, **kwargs)

    __code__ = __call__.__code__



assert all(
    '__slots__' in itp.__dict__
    for itp in AppItemType.__subclasses__()
)




def create_itemtype(tp_def: _parsing.ItemDefinition) -> type[AppItemType]:
    # XXX: Maybe move to `items.py`? It's the only caller.
    # Although it's kinda nice having this and the necessary
    # implementations in the same place.
    tp_name: str = tp_def.name # type: ignore
    reg_key = f'mvAppItemType::{tp_name}'

    # cached?
    if reg_key in AppItemMeta.__itemtype_registry__:
        cls = AppItemMeta.__itemtype_registry__[reg_key]

    else:
        params = inspect.signature(tp_def.command1).parameters

        ll_bases = []
        if "parent" not in params:
            if any(
                tp_def.command1.__name__.endswith(n) for n in ('_registry', 'theme')
            ):
                ll_bases.append(RegistryType)
            else:
                ll_bases.append(RootType)
        elif tp_def.is_container:
            ll_bases.append(ContainerType)
            ll_bases.append(BasicType)
        else:
            ll_bases.append(BasicType)

        hl_bases = []
        if tp_name == 'mvTable':
            hl_bases.append(TableItemType)
        elif tp_name.startswith("mvTable"):
            hl_bases.append(TableType)
        elif "Handler" in tp_name and not RegistryType in ll_bases:
            hl_bases.append(HandlerType)
        elif "Window" in tp_name:
            assert any(b in ll_bases for b in (ContainerType, RootType))
            hl_bases.append(WindowType)
        elif tp_name == 'mvDrawNode':
            hl_bases.append(DrawNodeType)
        elif tp_name.startswith('mvDraw') or "Drawlist" in tp_name:
            hl_bases.append(DrawingType)
        elif tp_name == 'mvFont':
            hl_bases.append(FontType)
        elif tp_name == "mvPlot":
            assert tp_def.is_container
            hl_bases.append(PlotType)
        elif tp_name == 'mvPlotAxis':
            hl_bases.append(PlotAxisType)
        elif any(n in tp_name for n in ("Plot", "Series", "mvAnnotation", "mvLegend")):
            hl_bases.append(PlottingType)
        elif tp_name == 'mvNodeEditor':
            hl_bases.append(NodeEditorType)
        elif tp_name.startswith("mvNode"):
            hl_bases.append(NodeType)
        elif tp_name == "mvTheme":
            hl_bases.append(ThemeType)
        elif tp_name.startswith("mvTheme"):
            hl_bases.append(ThemeElementType)

        fn_bases = []
        if any(
            s in tp_def.command1.__name__
            for s in ("_series", "theme_color", "theme_style")
        ):
            fn_bases.append(SupportsValueArray)
        if all(p in params for p in ("pos", "width", "height")):
            fn_bases.append(SupportsSized)
        if "callback" in params:
            fn_bases.append(SupportsCallback)

        bases = [*fn_bases, *hl_bases, *ll_bases]
        class_body = {
            '__doc__'  : _parsing.upd_command_doc(tp_def.command1),
            '__slots__': (),
        }

        cls = type(
            tp_name,
            tuple(bases),
            class_body,
            command=tp_def.command1,
            identity=(tp_def.enum, f"mvAppItemType::{tp_name}")
        )

        assert bases
        for b in cls.mro():
            if b in (object, int):
                continue
            assert hasattr(b, '__slots__')

    return cls  # type: ignore








# [ PUBLIC FUNCTIONS ]

@_exported
def interface(item: Item, *, initialize: bool = False, default_itp: type[_ITP] = mvAll) -> AppItemType | _ITP:
    """Return a proper `AppItemType` interface for an existing item
    based on its' internal type.

    Args:
        * item: Reference to an existing item.

        * initialize: If True, the interface's `__init__` method will
        be called during its creation. Default is False.

        * default_itp: Used as a fallback if a more specific interface
        type for the item cannot be found.
    """
    itp = AppItemMeta.__itemtype_registry__.get(
        ItemAPI.information(item)["type"],
        default_itp,
    )
    if initialize:
        return itp.new(item)
    return itp(tag=item)


@_exported
def registry_types():
    """Return all registry-type item classes."""
    return _get_itp_subclasses(RegistryType)


@_exported
def root_types():
    """Return all root-type item classes."""
    return _get_itp_subclasses(RootType)


@_exported
def container_types():
    """Return all container-type item classes."""
    return _get_itp_subclasses(ContainerType)


@_exported
def basic_types():
    """Return all non-container type item classes."""
    return _get_itp_subclasses(BasicType)


@_exported
def table_types():
    """Return all table-type item classes."""
    return _get_itp_subclasses(TableType)


@_exported
def drawing_items():
    """Return all drawing-type item classes."""
    return _get_itp_subclasses(DrawingType)


@_exported
def plotting_items():
    """Return all plotting-type item classes."""
    return _get_itp_subclasses(PlottingType)


@_exported
def node_types():
    """Return all node-type item classes."""
    return _get_itp_subclasses(NodeType)


@_exported
def theme_types():
    """Return all theming-type item classes."""
    return _get_itp_subclasses(ThemeType)


@overload
def auto_parent(parent_factory: ItemCommand, /) -> Callable[[type[_ITP]], type[_ITP]]: ...
@overload
def auto_parent(parent_factory: ItemCommand, /) -> Callable[[_T], _T]: ...
@overload
def auto_parent(parent_factory: ItemCommand, /, wrapped: type[_ITP]) -> type[_ITP]: ...
@overload
def auto_parent(parent_factory: ItemCommand, /, wrapped: _T) -> _T: ...
@_exported
def auto_parent(parent_factory: Any, /, wrapped: Any = None):  # type: ignore
    """Wrapper for `AppItemType` classes and functions that create
    Dear PyGui items. Before creating the item, a parent will
    automatically be created for the item if none is available.
        >>> # Usage w/interfaces
        >>> @auto_parent(dearpypixl.mvWindowAppItem)
        >>> class Frame(dearpypixl.mvChildWindow):
        ...     ...
        >>>
        >>> # Usage w/functions
        >>> @auto_parent(dearpygui.add_window)
        >>> def create_frame(*args, **kwargs):
        ...     ...
        ...     with dearpygui.child_window(*args, **kwargs) as frame:
        ...         ...
        ...     return frame

    Args:
        * parent_factory: A callable that, when called, creates a
        container item and returns a reference to it. Should be
        callable without arguments.

        * wrapped: A callable following the `ItemCommand` protocol.


    When wrapping `AppItemType` classes, this procedure specifically
    targets their `__init__` method.
    """
    def capture_wrapped(wrapped: Any):

        if isinstance(wrapped, type) and issubclass(wrapped, AppItemType):

            @functools.wraps(wrapped.__init__)
            def __init__(self, *args, **kwargs):
                if not kwargs.get('parent', 0) and not RegistryAPI.top_stack_item():
                    kwargs["parent"] = parent_factory()
                cls_init(self, *args, **kwargs)

            if wrapped.is_root_item:
                raise TypeError(
                    f"cannot set up auto parenting on a root interface "
                    f"type {wrapped.__qualname__!r}."
                )
            cls_init = wrapped.__init__
            wrapped.__init__ = __init__
            return wrapped

        elif callable(wrapped):

            @functools.wraps(wrapped)
            def item_command(*args, **kwargs):
                if not kwargs.get('parent', 0) and not RegistryAPI.top_stack_item():
                    kwargs["parent"] = parent_factory()
                return wrapped(*args, **kwargs)

            return item_command

        raise TypeError(
            f"expected a callable that creates items "
            f"for `wrapped`, got {type(wrapped).__qualname__!r}."
        )

    if wrapped is None:
        return capture_wrapped
    return capture_wrapped(wrapped)
